#
# Copyright (c) 2015-2016 CNRS
#

import numpy as np

from . import pinocchio_pywrap_default as pin
from .robot_wrapper import RobotWrapper


class RomeoWrapper(RobotWrapper):
    def __init__(self, filename, package_dirs=None, verbose=False):
        self.initFromURDF(
            filename,
            package_dirs=package_dirs,
            root_joint=pin.JointModelFreeFlyer(),
            verbose=verbose,
        )
        self.q0 = np.array(
            [
                0,
                0,
                0.840252,
                0,
                0,
                0,
                1,  # Free flyer
                0,
                0,
                -0.3490658,
                0.6981317,
                -0.3490658,
                0,  # left leg
                0,
                0,
                -0.3490658,
                0.6981317,
                -0.3490658,
                0,  # right leg
                0,  # chest
                1.5,
                0.6,
                -0.5,
                -1.05,
                -0.4,
                -0.3,
                -0.2,  # left arm
                0,
                0,
                0,
                0,  # head
                1.5,
                -0.6,
                0.5,
                1.05,
                -0.4,
                -0.3,
                -0.2,  # right arm
            ]
        )

        self.opCorrespondances = {
            "lh": "LWristPitch",
            "rh": "RWristPitch",
            "rf": "RAnkleRoll",
            "lf": "LAnkleRoll",
        }

        for op, name in self.opCorrespondances.items():
            self.__dict__[op] = self.index(name)
            # self.__dict__["_M" + op] = types.MethodType(
            # lambda s, q: s.position(q, idx), self
            # )

    # --- SHORTCUTS ---
    def Mrh(self, q):
        return self.position(q, self.rh)

    def Jrh(self, q):
        return self.jacobian(q, self.rh)

    def wJrh(self, q):
        return pin.jacobian(self.model, self.data, self.rh, q, False)

    def vrh(self, q, v):
        return self.velocity(q, v, self.rh)

    def Jlh(self, q):
        return self.jacobian(q, self.lh)

    def Mlh(self, q):
        return self.position(q, self.lh)

    def Jlf(self, q):
        return self.jacobian(q, self.lf)

    def Mlf(self, q):
        return self.position(q, self.lf)

    def Jrf(self, q):
        return self.jacobian(q, self.rf)

    def Mrf(self, q):
        return self.position(q, self.rf)


__all__ = ["RomeoWrapper"]



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import rand, zero
from test_case import PinocchioTestCase as TestCase

# common quantities for all tests.
# They correspond to the default values of the arguments, and they need to stay this way
r_coeff = 0.0
inv_damping = 0.0


class TestDynamicsBindings(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = rand(self.model.nv)
        self.tau = rand(self.model.nv)

        self.v0 = zero(self.model.nv)
        self.tau0 = zero(self.model.nv)
        self.tolerance = 1e-9

        # we compute J on a different self.data
        self.J = pin.computeJointJacobian(
            self.model,
            self.model.createData(),
            self.q,
            self.model.getJointId("lleg6_joint"),
        )
        self.gamma = zero(6)

    def test_forwardDynamics_default(self):
        data_no_q = self.model.createData()

        self.model.gravity = pin.Motion.Zero()
        ddq = pin.forwardDynamics(
            self.model, self.data, self.q, self.v0, self.tau0, self.J, self.gamma
        )
        self.assertLess(np.linalg.norm(ddq), self.tolerance)

        KKT_inverse = pin.getKKTContactDynamicMatrixInverse(
            self.model, self.data, self.J
        )
        M = pin.crba(self.model, self.model.createData(), self.q)

        self.assertApprox(
            M, np.linalg.inv(KKT_inverse)[: self.model.nv, : self.model.nv]
        )

        pin.computeAllTerms(self.model, data_no_q, self.q, self.v0)
        ddq_no_q = pin.forwardDynamics(
            self.model, data_no_q, self.tau0, self.J, self.gamma
        )
        self.assertLess(np.linalg.norm(ddq_no_q), self.tolerance)

        self.assertApprox(ddq, ddq_no_q)

    def test_computeKKTMatrix(self):
        model = self.model
        data = model.createData()
        data_ref = model.createData()

        q = self.q
        v = self.v
        tau = self.tau0
        J = self.J
        gamma = self.gamma

        pin.forwardDynamics(model, data_ref, q, v, tau, J, gamma)
        KKT_inverse_ref = pin.getKKTContactDynamicMatrixInverse(model, data_ref, J)

        KKT_inverse = pin.computeKKTContactDynamicMatrixInverse(model, data, q, J)
        KKT_inverse2 = pin.computeKKTContactDynamicMatrixInverse(model, data, q, J, 0.0)
        self.assertApprox(KKT_inverse, KKT_inverse_ref)
        self.assertApprox(KKT_inverse2, KKT_inverse_ref)

    def test_forwardDynamics_rcoeff(self):
        data_no_q = self.model.createData()

        self.model.gravity = pin.Motion.Zero()
        ddq = pin.forwardDynamics(
            self.model,
            self.data,
            self.q,
            self.v0,
            self.tau0,
            self.J,
            self.gamma,
            r_coeff,
        )
        self.assertLess(np.linalg.norm(ddq), self.tolerance)

        pin.computeAllTerms(self.model, data_no_q, self.q, self.v0)
        ddq_no_q = pin.forwardDynamics(
            self.model, data_no_q, self.tau0, self.J, self.gamma, r_coeff
        )
        self.assertLess(np.linalg.norm(ddq_no_q), self.tolerance)

        self.assertApprox(ddq, ddq_no_q)

    def test_forwardDynamics_q(self):
        data7 = self.data
        data8 = self.model.createData()
        _data9_deprecated = self.model.createData()
        ddq7 = pin.forwardDynamics(
            self.model, data7, self.q, self.v, self.tau, self.J, self.gamma
        )
        ddq8 = pin.forwardDynamics(
            self.model, data8, self.q, self.v, self.tau, self.J, self.gamma, r_coeff
        )

        self.assertTrue((ddq7 == ddq8).all())

    def test_forwardDynamics_no_q(self):
        data5 = self.data
        data6 = self.model.createData()
        data9_deprecated = self.model.createData()
        pin.computeAllTerms(self.model, data5, self.q, self.v0)
        pin.computeAllTerms(self.model, data6, self.q, self.v0)
        pin.computeAllTerms(self.model, data9_deprecated, self.q, self.v0)
        ddq5 = pin.forwardDynamics(self.model, data5, self.tau, self.J, self.gamma)
        ddq6 = pin.forwardDynamics(
            self.model, data6, self.tau, self.J, self.gamma, r_coeff
        )
        self.assertTrue((ddq5 == ddq6).all())

    def test_impulseDynamics_default(self):
        data_no_q = self.model.createData()

        vnext = pin.impulseDynamics(self.model, self.data, self.q, self.v0, self.J)
        self.assertLess(np.linalg.norm(vnext), self.tolerance)

        pin.crba(self.model, data_no_q, self.q)
        vnext_no_q = pin.impulseDynamics(self.model, data_no_q, self.v0, self.J)
        self.assertLess(np.linalg.norm(vnext_no_q), self.tolerance)

        self.assertApprox(vnext, vnext_no_q)

    def test_impulseDynamics_r(self):
        data_no_q = self.model.createData()

        vnext = pin.impulseDynamics(
            self.model, self.data, self.q, self.v0, self.J, r_coeff
        )
        self.assertLess(np.linalg.norm(vnext), self.tolerance)

        pin.crba(self.model, data_no_q, self.q)
        vnext_no_q = pin.impulseDynamics(
            self.model, data_no_q, self.v0, self.J, r_coeff
        )
        self.assertLess(np.linalg.norm(vnext_no_q), self.tolerance)

        self.assertApprox(vnext, vnext_no_q)

    def test_impulseDynamics_rd(self):
        data_no_q = self.model.createData()

        vnext = pin.impulseDynamics(
            self.model, self.data, self.q, self.v0, self.J, r_coeff, inv_damping
        )
        self.assertLess(np.linalg.norm(vnext), self.tolerance)

        pin.crba(self.model, data_no_q, self.q)
        vnext_no_q = pin.impulseDynamics(
            self.model, data_no_q, self.v0, self.J, r_coeff, inv_damping
        )
        self.assertLess(np.linalg.norm(vnext_no_q), self.tolerance)

        self.assertApprox(vnext, vnext_no_q)

    def test_impulseDynamics_q(self):
        data5 = self.data
        data6 = self.model.createData()
        data7 = self.model.createData()
        _data7_deprecated = self.model.createData()
        vnext5 = pin.impulseDynamics(self.model, data5, self.q, self.v, self.J)
        vnext6 = pin.impulseDynamics(self.model, data6, self.q, self.v, self.J, r_coeff)
        vnext7 = pin.impulseDynamics(
            self.model, data7, self.q, self.v, self.J, r_coeff, inv_damping
        )
        self.assertTrue((vnext5 == vnext6).all())
        self.assertTrue((vnext5 == vnext7).all())
        self.assertTrue((vnext6 == vnext7).all())

    def test_impulseDynamics_no_q(self):
        data4 = self.data
        data5 = self.model.createData()
        data6 = self.model.createData()
        data7_deprecated = self.model.createData()
        pin.crba(self.model, data4, self.q)
        pin.crba(self.model, data5, self.q)
        pin.crba(self.model, data6, self.q)
        pin.crba(self.model, data7_deprecated, self.q)
        vnext4 = pin.impulseDynamics(self.model, data4, self.v, self.J)
        vnext5 = pin.impulseDynamics(self.model, data5, self.v, self.J, r_coeff)
        vnext6 = pin.impulseDynamics(
            self.model, data6, self.v, self.J, r_coeff, inv_damping
        )

        self.assertTrue((vnext4 == vnext5).all())
        self.assertTrue((vnext4 == vnext6).all())
        self.assertTrue((vnext5 == vnext6).all())


if __name__ == "__main__":
    unittest.main()



"""
Ellipsoid Joint Kinematics Example

The joint has 3 DOF corresponding to 3 rotation angles that determine both:
- The position on the ellipsoid surface
- The orientation of the body

Note: the joint is not purely normal to the ellipsoid surface;
    it translates along a ellipsoid, but the rotational part is "normal" to a sphere.
"""

import numpy as np
import pinocchio as pin

# For visualization (optional)
try:
    from pinocchio.visualize import MeshcatVisualizer

    VISUALIZE = True
except ImportError:
    print("MeshcatVisualizer not available. Skipping visualization.")
    VISUALIZE = False


def create_ellipsoid_robot(radius_x=0.5, radius_y=0.3, radius_z=0.2):
    """
    Create a simple robot with an ellipsoid joint.

    Parameters
    ----------
    radius_x : float
        Ellipsoid radius along x-axis
    radius_y : float
        Ellipsoid radius along y-axis
    radius_z : float
        Ellipsoid radius along z-axis

    Returns
    -------
    model : pin.Model
        The robot model
    """
    model = pin.Model()

    # Create ellipsoid joint
    joint_name = "ellipsoid"
    joint_id = model.addJoint(
        0,  # parent joint id (0 = universe)
        pin.JointModelEllipsoid(radius_x, radius_y, radius_z),
        pin.SE3.Identity(),
        joint_name,
    )

    # Add a body with some inertia
    mass = 1.0
    lever = np.array([0.0, 0.0, 0.1])
    inertia = pin.Inertia(mass, lever, np.diag([0.01, 0.01, 0.01]))

    body_placement = pin.SE3.Identity()
    body_placement.translation = np.array([0.0, 0.0, 0.2])

    model.appendBodyToJoint(joint_id, inertia, body_placement)

    return model


def create_composite_model():
    """Create a model with a composite joint (6 DOF: Tx, Ty, Tz, Rx, Ry, Rz)."""
    model = pin.Model()

    # Create composite joint with 3 prismatic + 3 revolute
    jcomposite = pin.JointModelComposite()
    jcomposite.addJoint(pin.JointModelPX())
    jcomposite.addJoint(pin.JointModelPY())
    jcomposite.addJoint(pin.JointModelPZ())
    jcomposite.addJoint(pin.JointModelRX())
    jcomposite.addJoint(pin.JointModelRY())
    jcomposite.addJoint(pin.JointModelRZ())

    joint_id = model.addJoint(0, jcomposite, pin.SE3.Identity(), "composite")

    body_placement = pin.SE3.Identity()
    body_placement.translation = np.array([0.0, 0.0, 0.2])

    # Add a body with some inertia
    mass = 1.0
    lever = np.array([0.0, 0.0, 0.1])
    inertia = pin.Inertia(mass, lever, np.diag([0.01, 0.01, 0.01]))

    model.appendBodyToJoint(
        joint_id,
        inertia,
        body_placement,
    )

    return model


def compute_kinematics_example():
    """Demonstrate forward kinematics with the ellipsoid joint."""
    print("=" * 60)
    print("ELLIPSOID JOINT KINEMATICS EXAMPLE")
    print("=" * 60)

    # Create model with ellipsoid radii
    radius_x, radius_y, radius_z = 0.5, 0.3, 0.2
    model = create_ellipsoid_robot(radius_x, radius_y, radius_z)
    data = model.createData()

    print("\nEllipsoid parameters:")
    print(f"  radius_x (x-axis): {radius_x}")
    print(f"  radius_y (y-axis): {radius_y}")
    print(f"  radius_z (z-axis): {radius_z}")

    # Test different configurations
    configs = [
        ("Zero configuration", np.array([0.0, 0.0, 0.0])),
        ("Rotation around x", np.array([np.pi / 4, 0.0, 0.0])),
        ("Rotation around y", np.array([0.0, np.pi / 4, 0.0])),
        ("Rotation around z", np.array([0.0, 0.0, np.pi / 4])),
        ("Combined rotations", np.array([np.pi / 6, np.pi / 4, np.pi / 3])),
    ]

    for name, q in configs:
        pin.forwardKinematics(model, data, q)
        pin.updateFramePlacements(model, data)

        placement = data.oMi[1]
        position = placement.translation

        print(f"\n{name}:")
        print(f"  q = {q}")
        print(f"  Position on ellipsoid: {position}")
        print(f"  Distance from origin: {np.linalg.norm(position):.4f}")

        # Verify point is on ellipsoid surface
        normalized_pos = position / np.array([radius_x, radius_y, radius_z])

        # checking the quadratic form of the ellipsoid equation
        on_surface = np.abs(np.linalg.norm(normalized_pos) - 1.0) < 1e-10
        print(f"  On ellipsoid surface: {on_surface}")


def compute_dynamics_example():
    """Demonstrate dynamics computation with the ellipsoid joint."""
    print("\n" + "=" * 60)
    print("ELLIPSOID JOINT DYNAMICS EXAMPLE")
    print("=" * 60)

    # Create model
    model = create_ellipsoid_robot(0.5, 0.3, 0.2)
    data = model.createData()

    # Configuration, velocity, and acceleration
    q = np.array([np.pi / 6, np.pi / 4, np.pi / 3])
    v = np.array([0.1, 0.2, 0.3])
    a = np.array([0.01, 0.02, 0.03])

    pin.forwardKinematics(model, data, q, v, a)

    print(f"\nConfiguration: q = {q}")
    print(f"Velocity:      v = {v}")
    print(f"Acceleration:  a = {a}")

    # Compute dynamics with RNEA (Recursive Newton-Euler Algorithm)
    # Important Note:
    #   the generalized forces dimension matches the number of DOFs
    #   but they are not pure torques due to the joint's nature and because the last
    #   The last step of rnea is tau = S.T * f
    #   where S is the joint motion subspace, and f the spatial force of the joint
    #   So the resulting tau vector contains three generalized forces that
    #   include contributions from both rotational and translational effects.
    tau = pin.rnea(model, data, q, v, a)
    print(f"\nRequired torques (RNEA): τ = {tau}")

    # For example tau[0] is a combination of
    # S_11(q)* f_x + S_21(q)* f_y + S_31(q)* f_z
    #   + S_41(q)* τ_x + S_51(q)* τ_y + S_61(q)* τ_z

    # Create composite model for comparison of joint spatial forces `data.f[1]`
    composite_model = create_composite_model()
    composite_data = composite_model.createData()

    q_composite = np.zeros(6)
    q_composite[3:] = q  # Set only rotational part
    q_composite[:3] = data.oMi[
        1
    ].translation  # Set translation to current position on ellipsoid

    v_composite = np.zeros(6)
    v_composite[3:] = v
    v_composite[:3] = np.array(data.v[1])[:3]  # Set linear velocity part

    a_composite = np.zeros(6)
    a_composite[3:] = a
    a_composite[:3] = np.array(data.a[1])[:3]  # Set linear acceleration part

    tau_composite = pin.rnea(
        composite_model, composite_data, q_composite, v_composite, a_composite
    )
    print(f"\nComposite joint required torques (RNEA): τ = {tau_composite}")
    print("\nComparison of spatial forces at the joint:")
    print(f"  - Ellipsoid joint spatial force: f = {data.f[1]} ")
    print(f"  - Composite joint spatial force: f = {composite_data.f[1]} ")


def visualize_ellipsoid_motion():
    """Visualize the ellipsoid joint motion in MeshCat."""
    if not VISUALIZE:
        print("\nVisualization skipped (MeshcatVisualizer not available)")
        return

    if not hasattr(pin, "hppfcl"):
        print("\nVisualization skipped (hpp-fcl not available)")
        print("Install with: conda install -c conda-forge hpp-fcl")
        print("Then recompile Pinocchio with -DBUILD_WITH_HPP_FCL_SUPPORT=ON")
        return

    print("\n" + "=" * 60)
    print("ELLIPSOID JOINT VISUALIZATION")
    print("=" * 60)

    # Create model
    radius_x, radius_y, radius_z = 0.5, 0.3, 0.2
    model = create_ellipsoid_robot(radius_x, radius_y, radius_z)

    # Add visual geometry
    try:
        geom_model = pin.GeometryModel()

        # 1. Add the ELLIPSOID SURFACE as a visual object
        ellipsoid_shape = pin.hppfcl.Ellipsoid(radius_x, radius_y, radius_z)
        ellipsoid_geom = pin.GeometryObject(
            "ellipsoid_surface",
            0,  # Universe frame
            ellipsoid_shape,
            pin.SE3.Identity(),
        )
        ellipsoid_geom.meshColor = np.array(
            [0.8, 0.8, 0.8, 0.3]
        )  # Semi-transparent gray
        geom_model.addGeometryObject(ellipsoid_geom)

        # 2. Add a small sphere to show the contact point on the ellipsoid
        contact_sphere = pin.hppfcl.Sphere(0.03)
        contact_geom = pin.GeometryObject(
            "contact_point",
            model.getJointId("ellipsoid"),
            contact_sphere,
            pin.SE3.Identity(),  # At the joint frame (on the ellipsoid surface)
        )
        contact_geom.meshColor = np.array([1.0, 0.0, 0.0, 1.0])  # Red
        geom_model.addGeometryObject(contact_geom)

        # 3. Add a box to visualize the body orientation
        box = pin.hppfcl.Box(0.1, 0.1, 0.3)
        box_geom = pin.GeometryObject(
            "body_visual",
            model.getJointId("ellipsoid"),
            box,
            pin.SE3(np.eye(3), np.array([0.0, 0.0, 0.15])),
        )
        box_geom.meshColor = np.array([0.2, 0.6, 1.0, 1.0])  # Blue
        geom_model.addGeometryObject(box_geom)

        # Initialize visualizer
        viz = MeshcatVisualizer(model, geom_model, geom_model)
        viz.initViewer(open=True)
        viz.loadViewerModel()

        print("\n✓ Visualization initialized!")
        print("  - Gray ellipsoid: The constraint surface")
        print("  - Red sphere: Contact point on the surface")
        print("  - Blue box: The rigid body attached to the joint")
        print("\nAnimating ellipsoid joint motion...")
        print(
            "Open http://127.0.0.1:7000/static/ in your browser to see the animation."
        )

        # Animate through different configurations
        import time

        t = 0.0
        dt = 0.02

        for i in range(500):
            # Create a smooth trajectory on the ellipsoid
            q = np.array(
                [
                    0.8 * np.sin(1.0 * t),
                    0.6 * np.sin(1.2 * t + 1.0),
                    0.4 * np.sin(0.8 * t + 0.5),
                ]
            )

            viz.display(q)
            time.sleep(dt)
            t += dt

            if i % 50 == 0:
                print(
                    f"  Frame {i}/500 - Configuration: \n"
                    f"    [{q[0]:.3f}, {q[1]:.3f}, {q[2]:.3f}]"
                )

        # create extra motion that slides along the principal axes q0
        for angle in np.linspace(0, 2 * np.pi, 100):
            q = np.array([0.5 * np.cos(angle), 0.0, 0.0])
            viz.display(q)
            time.sleep(dt)

        # q1 axis
        for angle in np.linspace(0, 2 * np.pi, 100):
            q = np.array([0.0, 0.3 * np.cos(angle), 0.0])
            viz.display(q)
            time.sleep(dt)

        # q2 axis
        for angle in np.linspace(0, 2 * np.pi, 100):
            q = np.array([0.0, 0.0, 3.14 * np.cos(angle)])
            viz.display(q)
            time.sleep(dt)

        print("\n✓ Animation completed!")

    except Exception as e:
        print(f"\nVisualization error: {e}")
        import traceback

        traceback.print_exc()


def main():
    """Run all examples."""
    compute_kinematics_example()
    compute_dynamics_example()
    visualize_ellipsoid_motion()

    print("\n" + "=" * 60)
    print("Examples completed!")
    print("=" * 60)


if __name__ == "__main__":
    main()



from pathlib import Path

import numpy as np
import pinocchio

# Get URDF model
pinocchio_model_dir = Path(__file__).parent.parent / "models"
model_path = pinocchio_model_dir / "baxter_simple.urdf"
model_full = pinocchio.buildModelFromUrdf(model_path)

# To use the mimic tag from URDF set mimic flag to true
# (otherwise the tag will simply be ignored)
model_mimic_from_urdf = pinocchio.buildModelFromUrdf(model_path, mimic=True)

print(f"{model_full.nq=}")
print(f"{model_mimic_from_urdf.nq=}")

# Alternatively a "mimic" model can also be made by-hand
model_mimic = pinocchio.transformJointIntoMimic(model_full, 9, 10, -1.0, 0.0)
model_mimic = pinocchio.transformJointIntoMimic(model_mimic, 18, 19, -1.0, 0.0)

print(f"{(model_mimic_from_urdf == model_mimic)=}")  # True

# Creating manually the G matrix (cf. Featherstone's Rigid Body Dynamics Algorithms,
# chapter 10 about Gears if you want to know
# more about how joints mimic are handled here)
G = np.zeros([model_mimic.nv, model_full.nv])
for i in range(model_full.njoints):
    mimic_scaling = getattr(model_mimic.joints[i].extract(), "scaling", None)
    if mimic_scaling:
        G[model_mimic.joints[i].idx_v, model_full.joints[i].idx_v] = mimic_scaling
    else:
        # Not a mimic joint
        G[model_mimic.joints[i].idx_v, model_full.joints[i].idx_v] = 1
print("G = ")
print(np.array_str(G, precision=0, suppress_small=True, max_line_width=80))

# Random data
q_mimic = pinocchio.neutral(model_mimic)
v_mimic = np.random.random(model_mimic.nv)
a_mimic = np.random.random(model_mimic.nv)

# Expand configuration using the G matrix for the full model
# Possible because offset is 0, otherwise
# you need to add it to q of joints mimic
q_full = G.transpose() @ q_mimic

v_full = G.transpose() @ v_mimic
a_full = G.transpose() @ a_mimic

# Supported algorithms works as-is
data_mimic = model_mimic.createData()
data_full = model_full.createData()

tau_mimic = pinocchio.rnea(model_mimic, data_mimic, q_mimic, v_mimic, a_mimic)
tau_full = pinocchio.rnea(model_full, data_full, q_full, v_full, a_full)
print(f"{np.allclose(tau_mimic, G @ tau_full)=}")

M_mimic = pinocchio.crba(model_mimic, data_mimic, q_mimic)
M_full = pinocchio.crba(model_full, data_full, q_full)
print(f"{np.allclose(M_mimic, G @ M_full @ G.transpose())=}")

C_mimic = pinocchio.nle(model_mimic, data_mimic, q_mimic, v_mimic)
# For forward dynamics, aba does not support joints mimic.
# However it's still possible to compute acceleration of the
# mimic model, using the equation of motion tau = M*a + C
a_computed = np.linalg.solve(M_mimic, tau_mimic - C_mimic)
print(f"{np.allclose(a_mimic,a_computed)=}")



import unittest

import numpy as np
import pinocchio as pin


class TestModelGraphBindings(unittest.TestCase):
    def test_build_model(self):
        g = pin.graph.ModelGraph()
        g.addBody("body1", pin.Inertia.Identity())
        g.addBody("body2", pin.Inertia.Identity())

        g.edgeBuilder().withName("b1_b2").withJointType(
            pin.graph.JointPrismatic(np.array([1, 0, 0]))
        ).withSourceVertex("body1").withSourcePose(pin.SE3.Random()).withTargetVertex(
            "body2"
        ).withTargetPose(pin.SE3.Random()).build()

        m = pin.graph.buildModel(g, "body1", pin.SE3.Identity())
        self.assertTrue(m.njoints == 2)
        self.assertTrue(m.names[1] == "b1_b2")

        m, _ = pin.graph.buildModelWithBuildInfo(g, "body1", pin.SE3.Identity())
        self.assertTrue(m.njoints == 2)
        self.assertTrue(m.names[1] == "b1_b2")

    def test_merge(self):
        g1 = pin.graph.ModelGraph()
        g1.addBody("body1", pin.Inertia.Identity())
        g2 = pin.graph.ModelGraph()
        g2.addBody("body2", pin.Inertia.Identity())
        g = pin.graph.merge(
            g1,
            g2,
            "body1",
            "body2",
            pin.SE3.Random(),
            pin.graph.JointPrismatic(np.array([1, 0, 0])),
            "b1_b2",
        )
        m = pin.graph.buildModel(g, "body1", pin.SE3.Identity())
        self.assertTrue(m.njoints == 2)
        self.assertTrue(m.names[1] == "b1_b2")

    def test_lock_joints(self):
        g = pin.graph.ModelGraph()
        g.addBody("body1", pin.Inertia.Identity())
        g.addBody("body2", pin.Inertia.Identity())

        g.edgeBuilder().withName("b1_b2").withJointType(
            pin.graph.JointPrismatic(np.array([1, 0, 0]))
        ).withSourceVertex("body1").withSourcePose(pin.SE3.Random()).withTargetVertex(
            "body2"
        ).withTargetPose(pin.SE3.Random()).build()

        g_lock = pin.graph.lockJoints(g, ["b1_b2"], [np.array([0.3])])
        m = pin.graph.buildModel(g_lock, "body1", pin.SE3.Identity())
        self.assertTrue(m.njoints == 1)

    def test_prefix_names(self):
        g = pin.graph.ModelGraph()
        g.addBody("body1", pin.Inertia.Identity())
        g.addBody("body2", pin.Inertia.Identity())

        g.edgeBuilder().withName("b1_b2").withJointType(
            pin.graph.JointPrismatic(np.array([1, 0, 0]))
        ).withSourceVertex("body1").withSourcePose(pin.SE3.Random()).withTargetVertex(
            "body2"
        ).withTargetPose(pin.SE3.Random()).build()

        g1 = pin.graph.prefixNames(g, "g1/")
        m = pin.graph.buildModel(g1, "g1/body1", pin.SE3.Identity())
        self.assertTrue(m.njoints == 2)
        self.assertTrue(m.names[1] == "g1/b1_b2")

    def test_converter(self):
        g = pin.graph.ModelGraph()
        g.addBody("body1", pin.Inertia.Identity())
        g.addBody("body2", pin.Inertia.Identity())
        g.addBody("body3", pin.Inertia.Identity())

        g.edgeBuilder().withName("b1_b2").withJointType(
            pin.graph.JointPrismatic(np.array([1, 0, 0]))
        ).withSourceVertex("body1").withSourcePose(pin.SE3.Random()).withTargetVertex(
            "body2"
        ).withTargetPose(pin.SE3.Random()).build()

        g.edgeBuilder().withName("b2_b3").withJointType(
            pin.graph.JointPrismatic(np.array([0, 1, 0]))
        ).withSourceVertex("body2").withSourcePose(pin.SE3.Random()).withTargetVertex(
            "body3"
        ).withTargetPose(pin.SE3.Random()).build()

        m_body1, build_info_body1 = pin.graph.buildModelWithBuildInfo(
            g, "body1", pin.SE3.Identity()
        )
        q_body1 = pin.randomConfiguration(
            m_body1, np.array([-1.0] * 2), np.array([1.0] * 2)
        )
        v_body1 = np.random.random(m_body1.nv)

        m_body3, build_info_body3 = pin.graph.buildModelWithBuildInfo(
            g, "body3", pin.SE3.Identity()
        )
        converter = pin.graph.createConverter(
            m_body1, m_body3, build_info_body1, build_info_body3
        )
        q_body3 = converter.convertConfigurationVector(q_body1)
        self.assertTrue(q_body1[0] == -q_body3[-1])
        self.assertTrue(q_body1[1] == -q_body3[-2])

        v_body3 = converter.convertTangentVector(q_body1, v_body1)
        self.assertTrue(v_body1[0] == -v_body3[-1])
        self.assertTrue(v_body1[1] == -v_body3[-2])


if __name__ == "__main__":
    unittest.main()



import sys
import time
from pathlib import Path

import meshcat.geometry as g
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

# Load the URDF model.
pinocchio_model_dir = Path(__file__).parent.parent / "models"

model_path = pinocchio_model_dir / "example-robot-data/robots"
mesh_dir = pinocchio_model_dir

urdf_filename = "panda.urdf"
urdf_model_path = model_path / "panda_description/urdf" / urdf_filename

robot, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir
)
data = robot.createData()

# Start a new MeshCat server and client.
viz = MeshcatVisualizer(robot, collision_model, visual_model)
try:
    viz.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

# # Load the robot in the viewer.
viz.loadViewerModel()
# Display a robot configuration.
q0 = (robot.upperPositionLimit.T + robot.lowerPositionLimit.T) / 2
horizon = 0.2  # seconds
frame = robot.getFrameId(
    robot.frames[-1].name
)  # for example the last frame of the robot
n_samples = 5
facet_dims = 2

# To have convex hull computation or just the points of the reachable workspace and then
# compute it with cgal.
convex = True

if convex:
    verts, faces = pin.reachableWorkspaceHull(
        robot, q0, horizon, frame, n_samples, facet_dims
    )
    verts = verts.T

else:
    try:
        from CGAL.CGAL_Alpha_wrap_3 import *  # noqa: F403
        from CGAL.CGAL_Kernel import *  # noqa: F403
        from CGAL.CGAL_Mesh_3 import *  # noqa: F403
        from CGAL.CGAL_Polyhedron_3 import Polyhedron_3
    except ModuleNotFoundError:
        print("To compute non convex Polytope CGAL library needs to be installed.")
        sys.exit(0)

    def vertex_to_tuple(v):
        return v.x(), v.y(), v.z()

    def halfedge_to_triangle(he):
        p1 = he
        p2 = p1.next()
        p3 = p2.next()
        return [
            vertex_to_tuple(p1.vertex().point()),
            vertex_to_tuple(p2.vertex().point()),
            vertex_to_tuple(p3.vertex().point()),
        ]

    def alpha_shape_with_cgal(coords, alpha=None):
        """
        Compute the alpha shape of a set of points. (Code thanks to A. Skuric)
        Retrieved from http://blog.thehumangeo.com/2014/05/12/drawing-boundaries-in-python/

        :param coords : Coordinates of points
        :param alpha: List of alpha values to influence the gooeyness of the border.
        Smaller numbers don't fall inward as much as larger numbers.
        Too large, and you lose everything!
        :return: Shapely.MultiPolygons which is the hull of the input set of points
        """
        if alpha is None:
            bbox_diag = np.linalg.norm(np.max(coords, 0) - np.min(coords, 0))
            alpha_value = bbox_diag / 5
        else:
            alpha_value = np.mean(alpha)
        # Convert to CGAL point
        points = [Point_3(pt[0], pt[1], pt[2]) for pt in coords]  # noqa: F405
        # Compute alpha shape
        Q = Polyhedron_3()
        _a = alpha_wrap_3(points, alpha_value, 0.01, Q)  # noqa: F405

        alpha_shape_vertices = np.array(
            [vertex_to_tuple(vertex.point()) for vertex in Q.vertices()]
        )
        alpha_shape_faces = np.array(
            [np.array(halfedge_to_triangle(face.halfedge())) for face in Q.facets()]
        )

        return alpha_shape_vertices, alpha_shape_faces

    verts = pin.reachableWorkspace(robot, q0, horizon, frame, n_samples, facet_dims)
    verts = verts.T

    alpha = 0.2
    verts, faces = alpha_shape_with_cgal(verts, alpha)
    verts = faces.reshape(-1, 3)
    faces = np.arange(len(verts)).reshape(-1, 3)

print("------------------- Display Vertex")


# meshcat triangulated mesh
poly = g.TriangularMeshGeometry(vertices=verts, faces=faces)
viz.viewer["poly"].set_object(
    poly, g.MeshBasicMaterial(color=0x000000, wireframe=True, linewidth=12, opacity=0.2)
)

while True:
    viz.display(q0)
    viz.viewer["poly"].set_object(
        poly,
        g.MeshBasicMaterial(color=0x000000, wireframe=True, linewidth=2, opacity=0.2),
    )

    time.sleep(1e-2)



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestABA(TestCase):
    def setUp(self):
        # Set up non mimic model
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)
        self.ddq = np.random.rand(self.model.nv)

        self.fext = []
        for _ in range(self.model.njoints):
            self.fext.append(pin.Force.Random())

        # Set up mimic model
        self.model_mimic = pin.buildSampleModelManipulator(True)
        self.data_mimic = self.model_mimic.createData()
        self.q_mimic = np.zeros((self.model_mimic.nq,))

    def test_aba(self):
        model = self.model
        ddq = pin.aba(self.model, self.data, self.q, self.v, self.ddq)

        null_fext = pin.StdVec_Force()
        for _ in range(model.njoints):
            null_fext.append(pin.Force.Zero())

        ddq_null_fext = pin.aba(
            self.model, self.data, self.q, self.v, self.ddq, null_fext
        )
        self.assertApprox(ddq_null_fext, ddq)

        null_fext_list = []
        for f in null_fext:
            null_fext_list.append(f)

        print("size:", len(null_fext_list))
        ddq_null_fext_list = pin.aba(
            self.model, self.data, self.q, self.v, self.ddq, null_fext_list
        )
        self.assertApprox(ddq_null_fext_list, ddq)

    def test_computeMinverse(self):
        model = self.model
        Minv = pin.computeMinverse(model, self.data, self.q)

        data2 = model.createData()
        M = pin.crba(model, data2, self.q)

        self.assertApprox(np.linalg.inv(M), Minv)

    def test_assert_mimic_not_supported_function(self):
        self.assertRaises(
            RuntimeError,
            pin.computeMinverse,
            self.model_mimic,
            self.data_mimic,
            self.q_mimic,
        )


if __name__ == "__main__":
    unittest.main()



"""
Create a simulation environment for a N-pendulum.
Example of use:

env = Pendulum(N)
env.reset()

for i in range(1000):
   env.step(zero(env.nu))
   env.render()

"""

import time

import numpy as np
import pinocchio as pin
from display import Display
from numpy.linalg import inv
from pinocchio.utils import rand


class Visual:
    """
    Class representing one 3D mesh of the robot, to be attached to a joint. The class
    contains:
    * the name of the 3D objects inside Gepetto viewer.
    * the ID of the joint in the kinematic tree to which the body is attached.
    * the placement of the body with respect to the joint frame.
    This class is only used in the list Robot.visuals (see below).
    """

    def __init__(self, name, jointParent, placement):
        self.name = name  # Name in gepetto viewer
        self.jointParent = jointParent  # ID (int) of the joint
        self.placement = placement  # placement of the body wrt joint, i.e. bodyMjoint

    def place(self, display, oMjoint):
        oMbody = oMjoint * self.placement
        display.place(self.name, oMbody, False)


class Pendulum:
    """
    Define a class Robot with 7DOF (shoulder=3 + elbow=1 + wrist=3).
    The configuration is nq=7. The velocity is the same.
    The members of the class are:
    * viewer: a display encapsulating a gepetto viewer client to create 3D objects and
      place them.
    * model: the kinematic tree of the robot.
    * data: the temporary variables to be used by the kinematic algorithms.
    * visuals: the list of all the 'visual' 3D objects to render the robot, each element
      of the list being an object Visual (see above).

    See tp1.py for an example of use.
    """

    def __init__(self, nbJoint=1):
        """Create a Pinocchio model of a N-pendulum, with N the argument <nbJoint>."""
        self.viewer = Display()
        self.visuals = []
        self.model = pin.Model()
        self.createPendulum(nbJoint)
        self.data = self.model.createData()

        self.q0 = np.zeros(self.model.nq)

        self.DT = 5e-2  # Step length
        self.NDT = 2  # Number of Euler steps per integration (internal)
        self.Kf = 0.10  # Friction coefficient
        self.vmax = 8.0  # Max velocity (clipped if larger)
        self.umax = 2.0  # Max torque   (clipped if larger)
        self.withSinCos = False  # If true, state is [cos(q),sin(q),qdot], else [q,qdot]

    def createPendulum(self, nbJoint, rootId=0, prefix="", jointPlacement=None):
        color = [red, green, blue, transparency] = [1, 1, 0.78, 1.0]
        colorred = [1.0, 0.0, 0.0, 1.0]

        jointId = rootId
        jointPlacement = (
            jointPlacement if jointPlacement is not None else pin.SE3.Identity()
        )
        length = 1.0
        mass = length
        inertia = pin.Inertia(
            mass,
            np.array([0.0, 0.0, length / 2]),
            mass / 5 * np.diagflat([1e-2, length**2, 1e-2]),
        )

        for i in range(nbJoint):
            istr = str(i)
            name = prefix + "joint" + istr
            jointName, _bodyName = [name + "_joint", name + "_body"]
            jointId = self.model.addJoint(
                jointId, pin.JointModelRY(), jointPlacement, jointName
            )
            self.model.appendBodyToJoint(jointId, inertia, pin.SE3.Identity())
            try:
                self.viewer.viewer.gui.addSphere(
                    "world/" + prefix + "sphere" + istr, 0.15, colorred
                )
            except:  # noqa: E722
                pass
            self.visuals.append(
                Visual("world/" + prefix + "sphere" + istr, jointId, pin.SE3.Identity())
            )
            try:
                self.viewer.viewer.gui.addCapsule(
                    "world/" + prefix + "arm" + istr, 0.1, 0.8 * length, color
                )
            except:  # noqa: E722
                pass
            self.visuals.append(
                Visual(
                    "world/" + prefix + "arm" + istr,
                    jointId,
                    pin.SE3(np.eye(3), np.array([0.0, 0.0, length / 2])),
                )
            )
            jointPlacement = pin.SE3(np.eye(3), np.array([0.0, 0.0, length]))

        self.model.addFrame(
            pin.Frame("tip", jointId, 0, jointPlacement, pin.FrameType.OP_FRAME)
        )

    def display(self, q):
        pin.forwardKinematics(self.model, self.data, q)
        for visual in self.visuals:
            visual.place(self.viewer, self.data.oMi[visual.jointParent])
        self.viewer.viewer.gui.refresh()

    @property
    def nq(self):
        return self.model.nq

    @property
    def nv(self):
        return self.model.nv

    @property
    def nx(self):
        return self.nq + self.nv

    @property
    def nobs(self):
        return self.nx + self.withSinCos

    @property
    def nu(self):
        return self.nv

    def reset(self, x0=None):
        if x0 is None:
            q0 = np.pi * (rand(self.nq) * 2 - 1)
            v0 = rand(self.nv) * 2 - 1
            x0 = np.vstack([q0, v0])
        assert len(x0) == self.nx
        self.x = x0.copy()
        self.r = 0.0
        return self.obs(self.x)

    def step(self, u):
        assert len(u) == self.nu
        _, self.r = self.dynamics(self.x, u)
        return self.obs(self.x), self.r

    def obs(self, x):
        if self.withSinCos:
            return np.vstack(
                [np.vstack([np.cos(qi), np.sin(qi)]) for qi in x[: self.nq]]
                + [x[self.nq :]],
            )
        else:
            return x.copy()

    def tip(self, q):
        """Return the altitude of pendulum tip"""
        pin.framesKinematics(self.model, self.data, q)
        return self.data.oMf[1].translation[2, 0]

    def dynamics(self, x, u, display=False):
        """
        Dynamic function: x,u -> xnext=f(x,y).
        Put the result in x (the initial value is destroyed).
        Also compute the cost of making this step.
        Return x for convenience along with the cost.
        """

        def modulePi(th):
            return (th + np.pi) % (2 * np.pi) - np.pi

        def sumsq(x):
            return np.sum(np.square(x))

        cost = 0.0
        q = modulePi(x[: self.nq])
        v = x[self.nq :]
        u = np.clip(np.reshape(np.array(u), [self.nu, 1]), -self.umax, self.umax)

        DT = self.DT / self.NDT
        for i in range(self.NDT):
            pin.computeAllTerms(self.model, self.data, q, v)
            M = self.data.M
            b = self.data.nle
            # tau = u-self.Kf*v
            a = inv(M) * (u - self.Kf * v - b)

            v += a * DT
            q += v * DT
            cost += (sumsq(q) + 1e-1 * sumsq(v) + 1e-3 * sumsq(u)) * DT

            if display:
                self.display(q)
                time.sleep(1e-4)

        x[: self.nq] = modulePi(q)
        x[self.nq :] = np.clip(v, -self.vmax, self.vmax)

        return x, -cost

    def render(self):
        q = self.x[: self.nq]
        self.display(q)
        time.sleep(self.DT / 10)



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase


class TestKinematicsBindings(PinocchioTestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom(True, True)
        self.joint_idx = (
            self.model.getJointId("rarm2_joint")
            if self.model.existJointName("rarm2_joint")
            else (self.model.njoints - 1)
        )

    def test_getters(self):
        data = self.model.createData()
        q = pin.randomConfiguration(self.model)
        v = np.random.rand(self.model.nv)
        a = np.random.rand(self.model.nv)
        pin.forwardKinematics(self.model, data, q, v, a)

        T = data.oMi[self.joint_idx]

        v = pin.getVelocity(self.model, data, self.joint_idx)
        self.assertApprox(v, data.v[self.joint_idx])
        v = pin.getVelocity(self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL)
        self.assertApprox(v, data.v[self.joint_idx])
        v = pin.getVelocity(self.model, data, self.joint_idx, pin.ReferenceFrame.WORLD)
        self.assertApprox(v, data.oMi[self.joint_idx].act(data.v[self.joint_idx]))
        v = pin.getVelocity(
            self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )
        self.assertApprox(
            v, pin.SE3(T.rotation, np.zeros(3)).act(data.v[self.joint_idx])
        )

        a = pin.getAcceleration(self.model, data, self.joint_idx)
        self.assertApprox(a, data.a[self.joint_idx])
        a = pin.getAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL
        )
        self.assertApprox(a, data.a[self.joint_idx])
        a = pin.getAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.WORLD
        )
        self.assertApprox(a, data.oMi[self.joint_idx].act(data.a[self.joint_idx]))
        a = pin.getAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )
        self.assertApprox(
            a, pin.SE3(T.rotation, np.zeros(3)).act(data.a[self.joint_idx])
        )

        a = pin.getClassicalAcceleration(self.model, data, self.joint_idx)
        a = pin.getClassicalAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL
        )
        a = pin.getClassicalAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.WORLD
        )
        a = pin.getClassicalAcceleration(
            self.model, data, self.joint_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )


if __name__ == "__main__":
    unittest.main()



from pathlib import Path

import pinocchio as pin

pinocchio_model_dir = Path(__file__).parent.parent / "models"
urdf_filename = (
    pinocchio_model_dir
    / "example-robot-data/robots/anymal_b_simple_description/robots/anymal.urdf"
)
model = pin.buildModelFromUrdf(urdf_filename)
data = model.createData()

q0 = pin.neutral(model)

# Add feet frames
feet_names = ["LH_FOOT", "RH_FOOT", "LF_FOOT", "RF_FOOT"]
feet_frame_ids = []
for foot_name in feet_names:
    frame_id = model.getFrameId(foot_name)
    feet_frame_ids.append(frame_id)

contact_models = []
for fid in feet_frame_ids:
    frame = model.frames[fid]
    cmodel = pin.RigidConstraintModel(
        pin.ContactType.CONTACT_3D,
        frame.parent,
        frame.placement,
        pin.LOCAL_WORLD_ALIGNED,
    )
    contact_models.append(cmodel)

contact_data = [cmodel.createData() for cmodel in contact_models]

pin.initConstraintDynamics(model, data, contact_models)
pin.crba(model, data, q0)

data.contact_chol.compute(model, data, contact_models, contact_data)

delassus_matrix = data.contact_chol.getInverseOperationalSpaceInertiaMatrix()
delassus_matrix_inv = data.contact_chol.getOperationalSpaceInertiaMatrix()



import unittest

from pinocchio.utils import isapprox


def tracefunc(frame, event, arg):
    print(f"{event}, {frame.f_code.co_filename}: {frame.f_lineno}")
    return tracefunc


class PinocchioTestCase(unittest.TestCase):
    def assertApprox(self, a, b, eps=1e-6):
        return self.assertTrue(
            isapprox(a, b, eps),
            f"\n{a}\nis not approximately equal to\n{b}\nwith precision {eps:f}",
        )



import pickle
import unittest
from pathlib import Path

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestStdMap(TestCase):
    def setUp(self):
        pass

    def test_pickle(self):
        vec = pin.StdVec_Vector3()
        for k in range(100):
            vec.append(np.random.rand(3))

        pickle.dump(vec, Path("save_std_vec.p").open("wb"))

        vec_loaded = pickle.load(Path("save_std_vec.p").open("rb"))
        for k in range(len(vec)):
            self.assertApprox(vec[k], vec_loaded[k])


if __name__ == "__main__":
    unittest.main()



import unittest

import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestSampleModels(TestCase):
    def setUp(self):
        pass

    def test_all_sampled_models(self):
        huamnoid_1 = pin.buildSampleModelHumanoidRandom()
        huamnoid_2 = pin.buildSampleModelHumanoidRandom(True)
        huamnoid_3 = pin.buildSampleModelHumanoidRandom(False)

        self.assertTrue(huamnoid_1 != huamnoid_2)
        self.assertTrue(huamnoid_1 != huamnoid_3)

        manipulator_1 = pin.buildSampleModelManipulator()

        if pin.WITH_HPP_FCL:
            _geometry_manipulator_1 = pin.buildSampleGeometryModelManipulator(
                manipulator_1
            )

        humanoid_4 = pin.buildSampleModelHumanoid()
        humanoid_5 = pin.buildSampleModelHumanoid(True)
        humanoid_6 = pin.buildSampleModelHumanoid(False)

        self.assertTrue(humanoid_4 == humanoid_5)
        self.assertTrue(humanoid_4 != humanoid_6)

        if pin.WITH_HPP_FCL:
            _geometry_humanoid_2 = pin.buildSampleGeometryModelHumanoid(humanoid_4)


if __name__ == "__main__":
    unittest.main()



from time import sleep

import example_robot_data
import numpy as np
import pinocchio
from pinocchio.visualize import GepettoVisualizer

robot = example_robot_data.load("talos")
model = robot.model
data = robot.data

state_name = "half_sitting"

robot.q0 = robot.model.referenceConfigurations[state_name]

pinocchio.forwardKinematics(model, data, robot.q0)

lfFoot, rfFoot, lhFoot, rhFoot = (
    "left_sole_link",
    "right_sole_link",
    "gripper_left_fingertip_3_link",
    "gripper_right_fingertip_3_link",
)

foot_frames = [lfFoot, rfFoot, lhFoot, rhFoot]
foot_frame_ids = [robot.model.getFrameId(frame_name) for frame_name in foot_frames]
foot_joint_ids = [
    robot.model.frames[robot.model.getFrameId(frame_name)].parent
    for frame_name in foot_frames
]
pinocchio.forwardKinematics(model, data, robot.q0)
pinocchio.framesForwardKinematics(model, data, robot.q0)

constraint_models = []

for j, frame_id in enumerate(foot_frame_ids):
    contact_model_lf1 = pinocchio.RigidConstraintModel(
        pinocchio.ContactType.CONTACT_6D,
        robot.model,
        foot_joint_ids[j],
        robot.model.frames[frame_id].placement,
        0,
        data.oMf[frame_id],
    )
    constraint_models.extend([contact_model_lf1])

# Change arm position
constraint_models[3].joint2_placement = pinocchio.SE3(
    pinocchio.rpy.rpyToMatrix(np.array([0.0, -np.pi / 2, 0.0])),
    np.array([0.6, -0.40, 1.0]),
)

constraint_models[2].joint2_placement = pinocchio.SE3(
    pinocchio.rpy.rpyToMatrix(np.array([0, -np.pi / 2, 0.0])), np.array([0.6, 0.4, 1.0])
)

robot.setVisualizer(GepettoVisualizer())
robot.initViewer()
robot.loadViewerModel("pinocchio")
gui = robot.viewer.gui
robot.display(robot.q0)
window_id = robot.viewer.gui.getWindowID("python-pinocchio")

robot.viewer.gui.setBackgroundColor1(window_id, [1.0, 1.0, 1.0, 1.0])
robot.viewer.gui.setBackgroundColor2(window_id, [1.0, 1.0, 1.0, 1.0])
robot.viewer.gui.addFloor("hpp-gui/floor")

robot.viewer.gui.setScale("hpp-gui/floor", [0.5, 0.5, 0.5])
robot.viewer.gui.setColor("hpp-gui/floor", [0.7, 0.7, 0.7, 1.0])
robot.viewer.gui.setLightingMode("hpp-gui/floor", "OFF")

robot.display(robot.q0)

constraint_datas = [cm.createData() for cm in constraint_models]

q = robot.q0.copy()

pinocchio.computeAllTerms(model, data, q, np.zeros(model.nv))
kkt_constraint = pinocchio.ContactCholeskyDecomposition(model, constraint_models)
constraint_dim = sum([cm.size() for cm in constraint_models])
N = 100000
eps = 1e-10
mu = 1e-8  # 0.
# q_sol = (q[:] + np.pi) % np.pi - np.pi
q_sol = q.copy()
robot.display(q_sol)

# Bring CoM between the two feet.

mass = data.mass[0]


def squashing(model, data, q_in):
    q = q_in.copy()
    y = np.ones(constraint_dim)

    N_full = 200

    # Decrease CoMz by 0.2
    com_drop_amp = 0.1
    pinocchio.computeAllTerms(model, data, q, np.zeros(model.nv))
    com_base = data.com[0].copy()
    kp = 1.0
    speed = 1.0

    def com_des(k):
        return com_base - np.array(
            [
                0.0,
                0.0,
                np.abs(com_drop_amp * np.sin(2.0 * np.pi * k * speed / (N_full))),
            ]
        )

    for k in range(N):
        pinocchio.computeAllTerms(model, data, q, np.zeros(model.nv))
        pinocchio.computeJointJacobians(model, data, q)
        pinocchio.computeJointJacobians(model, data, q)
        com_act = data.com[0].copy()
        com_err = com_act - com_des(k)
        kkt_constraint.compute(model, data, constraint_models, constraint_datas, mu)
        constraint_value = np.concatenate(
            [pinocchio.log6(cd.c1Mc2) for cd in constraint_datas]
        )
        J = np.vstack(
            [
                pinocchio.getFrameJacobian(
                    model, data, cm.joint1_id, cm.joint1_placement, cm.reference_frame
                )
                for cm in constraint_models
            ]
        )
        primal_feas = np.linalg.norm(constraint_value, np.inf)
        print(J.shape, constraint_value.shape, y.shape)
        dual_feas = np.linalg.norm(J.T.dot(constraint_value + y), np.inf)
        print("primal_feas:", primal_feas)
        print("dual_feas:", dual_feas)
        # if primal_feas < eps and dual_feas < eps:
        #    print("Convergence achieved")
        #    break
        print("constraint_value:", np.linalg.norm(constraint_value))
        print("com_error:", np.linalg.norm(com_err))
        rhs = np.concatenate(
            [-constraint_value - y * mu, kp * mass * com_err, np.zeros(model.nv - 3)]
        )
        dz = kkt_constraint.solve(rhs)
        dy = dz[:constraint_dim]
        dq = dz[constraint_dim:]
        alpha = 1.0
        q = pinocchio.integrate(model, q, -alpha * dq)
        y -= alpha * (-dy + y)
        robot.display(q)
        sleep(0.05)
    return q


q_new = squashing(model, data, robot.q0)



import time

import numpy as np
from numpy import pi
from pendulum import Pendulum

p = Pendulum(1)

NQ = 20  # Discretization steps for position
NV = 19  # Discretization steps for velocity
VMAX = 5  # Max velocity (v in [-vmax,vmax])
NU = 9  # Discretization steps for torque
UMAX = 5  # Max torque (u in [-umax,umax])
DT = 3e-1

DQ = 2 * pi / NQ
DV = 2.0 * (VMAX) / NV
DU = 2.0 * (UMAX) / NU


# Continuous to discrete
def c2dq(q):
    q = (q + pi) % (2 * pi)
    return round(q / DQ) % NQ


def c2dv(v):
    v = np.clip(v, -VMAX + 1e-3, VMAX - 1e-3)
    return int(np.floor((v + VMAX) / DV))


def c2du(u):
    u = np.clip(u, -UMAX + 1e-3, UMAX - 1e-3)
    return int(np.floor((u + UMAX) / DU))


def c2d(qv):
    """From continuous to discrete."""
    return c2dq(qv[0]), c2dv(qv[1])


# Discrete to continuous
def d2cq(iq):
    iq = np.clip(iq, 0, NQ - 1)
    return iq * DQ - pi


def d2cv(iv):
    iv = np.clip(iv, 0, NV - 1) - (NV - 1) / 2
    return iv * DV


def d2cu(iu):
    iu = np.clip(iu, 0, NU - 1) - (NU - 1) / 2
    return iu * DU


def d2c(iqv):
    """From discrete to continuous"""
    return d2cq(iqv[0]), d2cv(iqv[1])


def x2i(x):
    return x[0] + x[1] * NQ


def i2x(i):
    return [i % NQ, i / NQ]


# --- PENDULUM


class DPendulum:
    def __init__(self):
        self.pendulum = Pendulum(1)
        self.pendulum.DT = DT
        self.pendulum.NDT = 5

    @property
    def nqv(self):
        return [NQ, NV]

    @property
    def nx(self):
        return NQ * NV

    @property
    def nu(self):
        return NU

    @property
    def goal(self):
        return x2i(c2d([0.0, 0.0]))

    def reset(self, x=None):
        if x is None:
            x = [np.random.randint(0, NQ), np.random.randint(0, NV)]
        else:
            x = i2x(x)
        assert len(x) == 2
        self.x = x
        return x2i(self.x)

    def step(self, iu):
        self.x = self.dynamics(self.x, iu)
        reward = 1 if x2i(self.x) == self.goal else 0
        return x2i(self.x), reward

    def render(self):
        q = d2cq(self.x[0])
        self.pendulum.display(
            np.array(
                [
                    q,
                ]
            )
        )
        time.sleep(self.pendulum.DT)

    def dynamics(self, ix, iu):
        x = np.array(d2c(ix))
        u = d2cu(iu)

        self.xc, _ = self.pendulum.dynamics(x, u)
        return c2d(x.T.tolist()[0])


"""
env = DPendulum()

print env.reset(x2i([14,11]))
hq = []
hv = []
hqc = []
hvc = []
u = 0
for i in range(100):
    ix,r=env.step(u)
    q,v = i2x(ix)
    env.render()
    if d2cv(v)==0.0: u = MAXU-1 if u==0 else 0
    hq.append( d2cq(env.x[0]) )
    hv.append( d2cv(env.x[1]) )
    hqc.append( env.xc[0,0] )
    hvc.append( env.xc[1,0] )

"""

"""


EPS = 1e-3
q = 0.0
v = -VMAX
hq = []
hv = []
hiq = []
hiv = []
hqa = []
hva = []

while q<2*pi:
    hq.append(q)
    iq = c2dq(q)
    hiq.append(iq)
    hqa.append(d2cq(iq))
    q += EPS
while v<VMAX:
    iv = c2dv(v)
    hv.append(v)
    hiv.append(iv)
    hva.append(d2cv(iv))
    v += EPS




"""



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestJointsAlgo(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)

    def test_basic(self):
        model = self.model

        q_ones = np.ones(model.nq)
        self.assertFalse(pin.isNormalized(model, q_ones))
        self.assertFalse(pin.isNormalized(model, q_ones, 1e-8))
        self.assertTrue(pin.isNormalized(model, q_ones, 1e2))

        q_rand = np.random.rand(model.nq)
        q_rand = pin.normalize(model, q_rand)
        self.assertTrue(pin.isNormalized(model, q_rand))
        self.assertTrue(pin.isNormalized(model, q_rand, 1e-8))

        self.assertTrue(abs(np.linalg.norm(q_rand[3:7]) - 1.0) <= 1e-8)

        q_next = pin.integrate(model, self.q, np.zeros(model.nv))
        self.assertApprox(q_next, self.q)

        v_diff = pin.difference(model, self.q, q_next)
        self.assertApprox(v_diff, np.zeros(model.nv))

        q_next = pin.integrate(model, self.q, self.v)
        q_int = pin.interpolate(model, self.q, q_next, 0.5)

        self.assertApprox(q_int, q_int)

        value = pin.squaredDistance(model, self.q, self.q)
        self.assertTrue((value <= 1e-8).all())

        dist = pin.distance(model, self.q, self.q)
        self.assertTrue(dist <= 1e-8)

        q_neutral = pin.neutral(model)
        self.assertApprox(q_neutral, q_neutral)

        q_rand1 = pin.randomConfiguration(model)
        q_rand2 = pin.randomConfiguration(model, -np.ones(model.nq), np.ones(model.nq))

        self.assertTrue(pin.isSameConfiguration(model, self.q, self.q, 1e-8))

        self.assertFalse(pin.isSameConfiguration(model, q_rand1, q_rand2, 1e-8))

    def test_derivatives(self):
        model = self.model

        q = self.q
        v = self.v

        J0, J1 = pin.dIntegrate(model, q, v)
        res_0 = pin.dIntegrate(model, q, v, pin.ArgumentPosition.ARG0)
        res_1 = pin.dIntegrate(model, q, v, pin.ArgumentPosition.ARG1)

        self.assertApprox(J0, res_0)
        self.assertApprox(J1, res_1)

        q_next = pin.integrate(model, q, v)

        J0, J1 = pin.dDifference(model, q, q_next)
        res_0 = pin.dDifference(model, q, q_next, pin.ARG0)
        res_1 = pin.dDifference(model, q, q_next, pin.ARG1)

        self.assertApprox(J0, res_0)
        self.assertApprox(J1, res_1)

    def test_transport(self):
        model = self.model
        Jq, Jv = pin.dIntegrate(model, self.q, self.v)
        mat = np.random.randn(model.nv, 2)

        mat_transported_q = pin.dIntegrateTransport(
            model, self.q, self.v, mat, pin.ARG0
        )
        mat_transported_v = pin.dIntegrateTransport(
            model, self.q, self.v, mat, pin.ARG1
        )

        self.assertApprox(mat_transported_q, np.dot(Jq, mat))
        self.assertApprox(mat_transported_v, np.dot(Jv, mat))


if __name__ == "__main__":
    unittest.main()



import unittest

import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestVersion(TestCase):
    def test_version(self):
        print("version:", pin.__version__)
        print("raw_version:", pin.__raw_version__)
        self.assertTrue(pin.__version__)
        self.assertTrue(pin.__raw_version__)


if __name__ == "__main__":
    unittest.main()



# ruff: noqa: F401
from importlib.util import find_spec

from .base_visualizer import BaseVisualizer
from .gepetto_visualizer import GepettoVisualizer
from .meshcat_visualizer import MeshcatVisualizer
from .panda3d_visualizer import Panda3dVisualizer
from .rviz_visualizer import RVizVisualizer

if find_spec("hppfcl") is not None:
    from .viser_visualizer import ViserVisualizer

from .visualizers import Visualizer



import numpy as np
import pinocchio as pin

# This example show how to:
#  - Construct a kinematics chain with ModelGraph API
#  - Bias a joint
#  - Merge two kinematics chain into a kinematics tree
#  - Lock a joint in a particular configuration
#  - Create a Model with a custom root joint

g = pin.graph.ModelGraph()

# Adding bodies to the pin.graph.
# In pinocchio, since bodies are represented as frames,
# there's 2 ways to add a body to the graph :
#   - addBody, helper function
#   - addFrame, where you create the BodyFrame, yourself
g.addBody("body1", pin.Inertia.Identity())
g.addFrame("body2", pin.graph.BodyFrame(pin.Inertia.Identity()))
g.addFrame("body3", pin.graph.BodyFrame(pin.Inertia.Identity()))

# Adding a sensor to the pin.graph.
g.addFrame("sensor1", pin.graph.SensorFrame())

# Now we add the joints between every body/sensor we have in the pin.graph.
pose_b1_to_j1 = pin.SE3.Random()  # pose of joint j1 wrt body1
pose_j1_to_b2 = pin.SE3.Random()  # pose of body2 wrt joint j1

# There are 2 ways to add joints to the graph.
# This helper function, where you specify the basics of a joint only.
g.addJoint(
    "j1",
    pin.graph.JointRevolute(np.array([0.0, 0.0, 1.0])),
    "body1",
    pose_b1_to_j1,
    "body2",
    pose_j1_to_b2,
)

# j2 will be biased by 50cm.
pose_b2_to_j2 = pin.SE3.Random()  # pose of joint j2 wrt body2
pose_j2_to_b3 = pin.SE3.Random()  # pose of body3 wrt joint j2

# To add a joint with more details, such as limits, bias... the
# builder interface is necessary.
g.edgeBuilder().withName("j2").withJointType(
    pin.graph.JointPrismatic(np.array([1, 0, 0]))
).withSourceVertex("body2").withSourcePose(pose_b2_to_j2).withTargetVertex(
    "body3"
).withTargetPose(pose_j2_to_b3).withQref(np.array([0.5])).build()

# sensor1 is a sensor frame so it can only be linked to the others body
# via a fixed joint
pose_b1_s1 = pin.SE3.Random()
pose_s1 = pin.SE3.Random()

g.edgeBuilder().withName("b1_s1").withJointType(
    pin.graph.JointFixed()
).withSourceVertex("body1").withSourcePose(pose_b1_s1).withTargetVertex(
    "sensor1"
).withTargetPose(pose_s1).build()

# Now we can choose which body will be our root its position, and build the model
kinematics_chain_from_body1 = pin.graph.buildModel(g, "body1", pin.SE3.Identity())
print("Kinematics chain from body1:")
print(kinematics_chain_from_body1)

# To merge two model, we can create a new ModelGraph and merge it to the first one.
# To simplify the process, we will append g to g.
# Since all joints and frames should have an unique name,
# we will use pin.graph.prefixNames function.
g1 = pin.graph.prefixNames(g, "g1/")
g2 = pin.graph.prefixNames(g, "g2/")

# Then we will attach g2/body3 to g1/body2 with a spherical joint.
# Equivalent to appendModel (deprecated now)
g1_g2_merged = pin.graph.merge(
    g1, g2, "g1/body2", "g2/body3", pin.SE3.Random(), pin.graph.JointSpherical()
)

# We can then create our model with any body as a root.
kinematics_tree_from_g1_body1 = pin.graph.buildModel(
    g1_g2_merged, "g1/body1", pin.SE3.Identity()
)
print("Kinematics tree from g1/body1:")
print(kinematics_tree_from_g1_body1)

# To lock a joints we only have to provide his name and his reference configuration.
# We will lock g1/j1 at 0.3 rad.
# Equivalent to buildReducedModel (deprecated now)
g1_g2_merged_locked = pin.graph.lockJoints(g1_g2_merged, ["g1/j1"], [np.array([0.3])])

# We can then create the locked model.
# We will use g2/body2 as root with a FreeFlyer joint.
kinematics_tree_from_g2_body2 = pin.graph.buildModel(
    g1_g2_merged_locked, "g2/body2", pin.SE3.Identity(), pin.graph.JointFreeFlyer()
)
print("Kinematics tree with g1/j1 locked and FreeFlyer from g2/body2:")
print(kinematics_tree_from_g2_body2)



import unittest

import pinocchio as pin
from test_case import PinocchioTestCase as TestCase

main_path = "./serialization-data"


class TestSerialization(TestCase):
    def testTXT(self):
        model = pin.buildSampleModelHumanoidRandom()
        filename = main_path + "/model.txt"
        model.saveToText(filename)

        model2 = pin.Model()
        model2.loadFromText(filename)

        self.assertTrue(model == model2)

    def testXML(self):
        model = pin.buildSampleModelHumanoidRandom()
        filename = main_path + "/model.xml"
        tag_name = "Model"
        model.saveToXML(filename, tag_name)

        model2 = pin.Model()
        model2.loadFromXML(filename, tag_name)

        self.assertTrue(model == model2)

    def testBIN(self):
        model = pin.buildSampleModelHumanoidRandom()
        filename = main_path + "/model.bin"
        model.saveToBinary(filename)

        model2 = pin.Model()
        model2.loadFromBinary(filename)

        self.assertTrue(model == model2)


if __name__ == "__main__":
    unittest.main()



import math
import os
import sys
import time
from pathlib import Path

import hppfcl as fcl
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer as Visualizer

# load model from example-robot urdf
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "ur5_robot.urdf"
urdf_model_path = model_path / "ur_description/urdf" / urdf_filename

model1, collision_model1, visual_model1 = pin.buildModelsFromUrdf(
    urdf_model_path, package_dirs=mesh_dir
)

# build model from scratch
model2 = pin.Model()
model2.name = "pendulum"
geom_model = pin.GeometryModel()

parent_id = 0
joint_placement = pin.SE3.Identity()
body_mass = 1.0
body_radius = 1e-2

joint_name = "joint_spherical"
joint_id = model2.addJoint(
    parent_id, pin.JointModelSpherical(), joint_placement, joint_name
)

body_inertia = pin.Inertia.FromSphere(body_mass, body_radius)
body_placement = joint_placement.copy()
body_placement.translation[2] = 0.1
model2.appendBodyToJoint(joint_id, body_inertia, body_placement)

geom1_name = "ball"
shape1 = fcl.Sphere(body_radius)
geom1_obj = pin.GeometryObject(geom1_name, joint_id, body_placement, shape1)
geom1_obj.meshColor = np.ones(4)
geom_model.addGeometryObject(geom1_obj)

geom2_name = "bar"
shape2 = fcl.Cylinder(body_radius / 4.0, body_placement.translation[2])
shape2_placement = body_placement.copy()
shape2_placement.translation[2] /= 2.0

geom2_obj = pin.GeometryObject(geom2_name, joint_id, shape2_placement, shape2)
geom2_obj.meshColor = np.array([0.0, 0.0, 0.0, 1.0])
geom_model.addGeometryObject(geom2_obj)

visual_model2 = geom_model

# join the two models, append pendulum to end effector
frame_id_end_effector = model1.getFrameId("tool0")
model, visual_model = pin.appendModel(
    model1,
    model2,
    visual_model1,
    visual_model2,
    frame_id_end_effector,
    pin.SE3.Identity(),
)

print(
    f"Check the joints of the appended model:\n {model} \n "
    "->Notice the spherical joint at the end."
)

try:
    viz = Visualizer(model, visual_model, visual_model)
    viz.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

# Load the robot in the viewer.
viz.loadViewerModel()

# Display a random robot configuration.
model.lowerPositionLimit.fill(-math.pi / 2)
model.upperPositionLimit.fill(+math.pi / 2)
q = pin.randomConfiguration(model)
viz.display(q)
time.sleep(1.0)



"""
Pose a Solo-12 robot on a surface defined through a function and displayed through an
hppfcl.HeightField.
"""

import time
from pathlib import Path

import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

# Load the URDF model.
# Conversion with str seems to be necessary when executing this file with ipython
pinocchio_model_dir = Path(__file__).parent.parent / "models"

model_path = pinocchio_model_dir / "example-robot-data/robots"
mesh_dir = pinocchio_model_dir
urdf_filename = "solo12.urdf"
urdf_model_path = model_path / "solo_description/robots" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir, pin.JointModelFreeFlyer()
)

q_ref = np.array(
    [
        [0.09906518],
        [0.20099078],
        [0.32502457],
        [0.19414175],
        [-0.00524735],
        [-0.97855773],
        [0.06860185],
        [0.00968163],
        [0.60963582],
        [-1.61206407],
        [-0.02543309],
        [0.66709088],
        [-1.50870083],
        [0.32405118],
        [-1.15305599],
        [1.56867351],
        [-0.39097222],
        [-1.29675892],
        [1.39741073],
    ]
)


vizer = MeshcatVisualizer(model, collision_model, visual_model)
vizer.initViewer(open=True)


def ground(xy):
    return (
        np.sin(xy[0] * 3) / 5
        + np.cos(xy[1] ** 2 * 3) / 20
        + np.sin(xy[1] * xy[0] * 5) / 10
    )


def vizGround(viz, elevation_fn, space, name="ground", color=[1.0, 1.0, 0.6, 0.8]):
    xg = np.arange(-2, 2, space)
    nx = xg.shape[0]
    xy_g = np.meshgrid(xg, xg)
    xy_g = np.stack(xy_g)
    elev_g = np.zeros((nx, nx))
    elev_g[:, :] = elevation_fn(xy_g)

    sx = xg[-1] - xg[0]
    sy = xg[-1] - xg[0]
    elev_g[:, :] = elev_g[::-1, :]
    import hppfcl

    heightField = hppfcl.HeightFieldAABB(sx, sy, elev_g, np.min(elev_g))
    pl = pin.SE3.Identity()
    obj = pin.GeometryObject("ground", 0, pl, heightField)
    obj.meshColor[:] = color
    viz.addGeometryObject(obj)


# Load the robot in the viewer.
vizer.loadViewerModel()

colorrgb = [128, 149, 255, 200]
colorrgb = np.array(colorrgb) / 255.0
vizGround(vizer, ground, 0.02, color=colorrgb)

vizer.display(q_ref)
time.sleep(1.0)



from enum import Enum
from importlib.util import find_spec
from os import environ

from .base_visualizer import BaseVisualizer
from .gepetto_visualizer import GepettoVisualizer
from .meshcat_visualizer import MeshcatVisualizer
from .panda3d_visualizer import Panda3dVisualizer
from .rviz_visualizer import RVizVisualizer

if find_spec("hppfcl") is not None:
    from .viser_visualizer import ViserVisualizer
else:

    class ViserVisualizer:
        pass


class Visualizer(Enum):
    BASE = BaseVisualizer
    GEPETTO = GepettoVisualizer
    MESHCAT = MeshcatVisualizer
    PANDA3D = Panda3dVisualizer
    RVIZ = RVizVisualizer
    VISER = ViserVisualizer

    @classmethod
    def default(cls):
        """
        Allow user to choose their prefered viewer with eg.
        export PINOCCHIO_VIEWER=RVIZ.

        Otherwise, try to find one which is installed.
        """
        # Allow user to define which viewer they want
        if "PINOCCHIO_VIEWER" in environ:
            selected = environ["PINOCCHIO_VIEWER"].upper()
            if hasattr(cls, selected):
                return getattr(cls, selected).value
            err = (
                f"The visualizer '{selected}' is not available.\n"
                "Please set PINOCCHIO_VIEWER to something installed among:\n"
                "- meshcat\n"
                "- gepetto-viewer\n"
                "- panda3d\n"
                "- rviz\n"
                "- viser\n"
            )
            raise ImportError(err)

        # Otherwise, use the first available
        for v in ["meshcat", "gepetto", "panda3d_viewer", "rviz", "viser"]:
            if find_spec(v) is not None:
                return getattr(cls, v.replace("_viewer", "").upper()).value

        err = (
            "No visualizer could be found.\n"
            "Please install one of those:\n"
            "- meshcat\n"
            "- gepetto-viewer\n"
            "- panda3d\n"
            "- rviz\n"
            "- viser\n"
        )
        raise ImportError(err)



from sys import argv

import pinocchio as pin
from numpy import pi
from pinocchio.visualize import GepettoVisualizer, MeshcatVisualizer, RVizVisualizer

# GepettoVisualizer: -g
# MeshcatVisualizer: -m
VISUALIZER = None
if len(argv) > 1:
    opt = argv[1]
    if opt == "-g":
        VISUALIZER = GepettoVisualizer
    elif opt == "-m":
        VISUALIZER = MeshcatVisualizer
    elif opt == "-r":
        VISUALIZER = RVizVisualizer
    else:
        raise ValueError("Unrecognized option: " + opt)

model = pin.buildSampleModelHumanoid()
visual_model = pin.buildSampleGeometryModelHumanoid(model)
collision_model = visual_model.copy()

q0 = pin.neutral(model)

if VISUALIZER:
    viz = VISUALIZER(model, collision_model, visual_model)
    viz.initViewer()
    viz.loadViewerModel()
    viz.display(q0)

    input("Enter to check a new configuration")

    q = q0.copy()
    q[8] = pi / 2
    q[14] = pi / 2
    q[23] = -pi / 2
    q[29] = pi / 2

    viz.display(q)

    input("Press enter to exit...")



import os
import unittest
from pathlib import Path

import pinocchio as pin


@unittest.skipUnless(pin.WITH_URDFDOM, "Needs URDFDOM")
class TestGeometryObjectUrdfBindings(unittest.TestCase):
    def setUp(self):
        self.model_dir = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
        self.model_path = self.model_dir / "romeo_description/urdf/romeo.urdf"

    def test_load(self):
        pin.buildModelFromUrdf(self.model_path)
        pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())

    def test_self_load(self):
        model = pin.Model()
        pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer(), model)
        pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())

    def test_xml(self):
        with self.model_path.open() as model:
            file_content = model.read()

        model_ref = pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())
        model = pin.buildModelFromXML(file_content, pin.JointModelFreeFlyer())

        self.assertEqual(model, model_ref)

        model_self = pin.Model()
        pin.buildModelFromXML(file_content, pin.JointModelFreeFlyer(), model_self)
        self.assertEqual(model_self, model_ref)

    def test_pickle(self):
        import pickle

        model_path = self.model_dir / "ur_description/urdf/ur5_robot.urdf"

        model = pin.buildModelFromUrdf(model_path)
        filename = Path("model.pickle")
        with filename.open("wb") as f:
            pickle.dump(model, f)

        with filename.open("rb") as f:
            model_copy = pickle.load(f)

        self.assertTrue(model == model_copy)


if __name__ == "__main__":
    unittest.main()



class FootSteps:
    """
    The class stores three functions of time: left, right and flying_foot.
    Each function is piecewise constant. For each function, the user can ask
    what is the value of this function at time t.

    The storage is composed of three lists for left, right and flying_foot, and a list
    for time.
    The list of times stores the time intervals, i.e. each element of the list is
    the start of a time interval. The first element of the list is 0.
    The value of the functions left, right, flying_foot one this time interval is stored
    at the same position is their respective list (i.e. value of left on interval
    [time[i], time[i+1]] is stored in left[i].

    The 4 lists are set up using function add_phase().
    The values of functions left, right, flying_foot can be accessed through the
    function get_phase_type(t), get_left_position(t), get_right_position(t).  phase_type
    are 'left' (meaning left foot is flying, right foot is fixed), 'right' (ie the
    opposite) or 'none' (meaning no foot is flying, both are fixed on the ground).

    Additionnally, functions get_left_next_position(t),
    get_right_next_position(t) can be used to get the next position of the
    flying foot (in that case, additional work is needed to compute the
    position of flying foot at time t by interpolating get_left_position(t)
    and get_left_next_position(t).

    Functions get_phase_start(t), get_phase_duration(t) and get_phase_remaining(t)
    can be used to get the starting time, the duration and the remaining time of the
    current phase at time t.
    """

    def __init__(self, right, left):
        """
        The class is initiated from the initial positions of left and right feet.
        """
        self.right = [right]
        self.left = [left]
        self.time = [0.0]
        self.flying_foot = []

    def add_phase(self, duration, foot, position=None):
        """
        Add a phase lasting <duration> where the flyhing foot <foot> (either 'left' or
        'right') moves to <position> (being a vector or a SE3 placement).
        Alternatively, <foot> might be set to 'none' (i.e double support). In that case,
        <position> is not specified (or is set to None, default).
        """
        assert foot == "left" or foot == "right" or foot == "none"
        self.time.append(self.time[-1] + duration)
        self.right.append(self.right[-1])
        self.left.append(self.left[-1])
        self.flying_foot.append(foot)
        if foot == "left":
            self.left[-1] = position
        elif foot == "right":
            self.right[-1] = position

    def get_index_from_time(self, t):
        """
        Return the index i of the interval containing t, i.e. t in time[i], time[i+1]
        """
        if t > self.time[-1]:
            return len(self.time) - 1
        return next(i for i, ti in enumerate(self.time) if ti > t) - 1

    def get_phase_type(self, t):
        i = self.get_index_from_time(t)
        return self.flying_foot[i]

    def is_double_from_left_to_right(self, t):
        """
        Suppose that phase at time <t> is a double support phase.
        Return True if the previous phase is left and/or the next phase is right.
        """
        assert self.get_phase_type(t) == "none"
        i = self.get_index_from_time(t)
        return (
            self.flying_foot[i - 1] == "left"
            if i > 0
            else self.flying_foot[i + 1] == "right"
        )

    def get_left_position(self, t):
        return self.left[self.get_index_from_time(t)]

    def get_right_position(self, t):
        return self.right[self.get_index_from_time(t)]

    def get_left_next_position(self, t):
        i = self.get_index_from_time(t)
        i = i + 1 if i + 1 < len(self.time) else i
        return self.left[i]

    def get_right_next_position(self, t):
        i = self.get_index_from_time(t)
        i = i + 1 if i + 1 < len(self.time) else i
        return self.right[i]

    def get_phase_start(self, t):
        return self.time[self.get_index_from_time(t)]

    def get_phase_duration(self, t):
        i = self.get_index_from_time(t)
        return self.time[i + 1] - self.time[i]

    def get_phase_remaining(self, t):
        return self.time[self.get_index_from_time(t) + 1] - t



import pinocchio

model = pinocchio.buildSampleModelManipulator()
data = model.createData()

q = pinocchio.neutral(model)
v = pinocchio.utils.zero(model.nv)
a = pinocchio.utils.zero(model.nv)

tau = pinocchio.rnea(model, data, q, v, a)
print("tau = ", tau.T)



import functools
import warnings


class DeprecatedWarning(UserWarning):
    pass


def deprecated(instructions):
    """Flags a method as deprecated.
    Args:
        instructions: A human-friendly string of instructions, such
            as: 'Please migrate to add_proxy() ASAP.'
    """

    def decorator(func):
        """This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            message = f"Call to deprecated function {func.__name__}. {instructions}"

            warnings.warn(message, category=DeprecatedWarning, stacklevel=2)

            return func(*args, **kwargs)

        instructions_doc = "Deprecated: " + instructions
        if wrapper.__doc__ is None:
            wrapper.__doc__ = instructions_doc
        else:
            wrapper.__doc__ = wrapper.__doc__.rstrip() + "\n\n" + instructions_doc
        return wrapper

    return decorator



import numpy as np
import pinocchio as pin

# Create model and data

model = pin.buildSampleModelHumanoidRandom()
data = model.createData()

# Set bounds (by default they are undefinded)

model.lowerPositionLimit = -np.ones((model.nq, 1))
model.upperPositionLimit = np.ones((model.nq, 1))

q = pin.randomConfiguration(model)  # joint configuration
v = np.random.rand(model.nv, 1)  # joint velocity
a = np.random.rand(model.nv, 1)  # joint acceleration

# Evaluate all the terms required by the kinematics derivatives

pin.computeForwardKinematicsDerivatives(model, data, q, v, a)

# Evaluate the derivatives for a precise joint (e.g. rleg6_joint)

joint_name = "rleg6_joint"
joint_id = model.getJointId(joint_name)

# Derivatives of the spatial velocity with respect to the joint configuration and
# velocity vectors.

(dv_dq, dv_dv) = pin.getJointVelocityDerivatives(
    model, data, joint_id, pin.ReferenceFrame.WORLD
)
# or to get them in the LOCAL frame of the joint
(dv_dq_local, dv_dv_local) = pin.getJointVelocityDerivatives(
    model, data, joint_id, pin.ReferenceFrame.LOCAL
)

# Derivatives of the spatial acceleration of the joint with respect to the joint
# configuration, velocity and acceleration vectors.

(dv_dq, da_dq, da_dv, da_da) = pin.getJointAccelerationDerivatives(
    model, data, joint_id, pin.ReferenceFrame.WORLD
)
# or to get them in the LOCAL frame of the joint
(dv_dq_local, da_dq_local, da_dv_local, da_da_local) = (
    pin.getJointAccelerationDerivatives(model, data, joint_id, pin.ReferenceFrame.LOCAL)
)



#
# Copyright (c) 2016 CNRS
#

import numpy as np
import pinocchio as pin
from dcrba import DCRBA, DRNEA, Coriolis
from numpy.linalg import norm
from pinocchio.robot_wrapper import RobotWrapper
from pinocchio.utils import rand

np.random.seed(0)

robot = RobotWrapper(
    "/home/nmansard/src/pinocchio/pinocchio/models/romeo/urdf/romeo.urdf",
    [
        "/home/nmansard/src/pinocchio/pinocchio/models/romeo/",
    ],
    pin.JointModelFreeFlyer(),
)
q = rand(robot.model.nq)
q[3:7] /= norm(q[3:7])
vq = rand(robot.model.nv)
aq = rand(robot.model.nv)

# d/dq M(q)
dcrba = DCRBA(robot)
dcrba.pre(q)
Mp = dcrba()

# d/dvq RNEA(q,vq) = C(q,vq)
coriolis = Coriolis(robot)
C = coriolis(q, vq)

# d/dq RNEA(q,vq,aq)
drnea = DRNEA(robot)
aq = rand(robot.model.nv)
R = drnea(q, vq, aq)



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import rand
from test_case import PinocchioTestCase as TestCase


class TestComBindings(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)

    def test_mass(self):
        mass = pin.computeTotalMass(self.model)
        self.assertIsNot(mass, np.nan)

        mass_check = sum([inertia.mass for inertia in self.model.inertias[1:]])
        self.assertApprox(mass, mass_check)

        mass_data = pin.computeTotalMass(self.model, self.data)
        self.assertIsNot(mass_data, np.nan)
        self.assertApprox(mass, mass_data)
        self.assertApprox(mass_data, self.data.mass[0])

        data2 = self.model.createData()
        pin.centerOfMass(self.model, data2, self.q)
        self.assertApprox(mass, data2.mass[0])

    def test_subtree_masses(self):
        pin.computeSubtreeMasses(self.model, self.data)

        data2 = self.model.createData()
        pin.centerOfMass(self.model, data2, self.q)

        for i in range(self.model.njoints):
            self.assertApprox(self.data.mass[i], data2.mass[i])

    def test_com_0(self):
        data = self.data

        c0 = pin.centerOfMass(self.model, self.data, self.q)
        c0_bis = pin.centerOfMass(self.model, self.data, self.q, False)

        self.assertApprox(c0, c0_bis)

        data2 = self.model.createData()
        pin.forwardKinematics(self.model, data, self.q)
        c0 = pin.centerOfMass(self.model, data, pin.POSITION)
        pin.forwardKinematics(self.model, data2, self.q)
        c0_bis = pin.centerOfMass(self.model, data2, pin.POSITION, False)

        self.assertApprox(c0, c0_bis)

        c0_bis = pin.centerOfMass(self.model, self.data, 0)

        self.assertApprox(c0, c0_bis)

        self.assertApprox(c0, data2.com[0])
        self.assertApprox(self.data.com[0], data2.com[0])

    def test_com_1(self):
        data = self.data

        v = rand(self.model.nv)
        c0 = pin.centerOfMass(self.model, self.data, self.q, v)
        c0_bis = pin.centerOfMass(self.model, self.data, self.q, v, False)

        self.assertApprox(c0, c0_bis)

        data2 = self.model.createData()
        pin.forwardKinematics(self.model, data, self.q, v)
        c0 = pin.centerOfMass(self.model, data, pin.VELOCITY)
        pin.forwardKinematics(self.model, data2, self.q, v)
        c0_bis = pin.centerOfMass(self.model, data2, pin.VELOCITY, False)

        self.assertApprox(c0, c0_bis)

        c0_bis = pin.centerOfMass(self.model, data2, 1)

        self.assertApprox(c0, c0_bis)

        data3 = self.model.createData()
        pin.centerOfMass(self.model, data3, self.q)

        self.assertApprox(self.data.com[0], data2.com[0])
        self.assertApprox(self.data.vcom[0], data2.vcom[0])

        self.assertApprox(self.data.com[0], data3.com[0])

    def test_com_2(self):
        data = self.data

        v = rand(self.model.nv)
        a = rand(self.model.nv)
        c0 = pin.centerOfMass(self.model, self.data, self.q, v, a)
        c0_bis = pin.centerOfMass(self.model, self.data, self.q, v, a, False)

        self.assertApprox(c0, c0_bis)

        data2 = self.model.createData()
        pin.forwardKinematics(self.model, data, self.q, v, a)
        c0 = pin.centerOfMass(self.model, data, pin.ACCELERATION)
        pin.forwardKinematics(self.model, data2, self.q, v, a)
        c0_bis = pin.centerOfMass(self.model, data2, pin.ACCELERATION, False)

        self.assertApprox(c0, c0_bis)

        c0_bis = pin.centerOfMass(self.model, data2, 2)
        self.assertApprox(c0, c0_bis)

        data3 = self.model.createData()
        pin.centerOfMass(self.model, data3, self.q)

        data4 = self.model.createData()
        pin.centerOfMass(self.model, data4, self.q, v)

        self.assertApprox(self.data.com[0], data2.com[0])
        self.assertApprox(self.data.vcom[0], data2.vcom[0])
        self.assertApprox(self.data.acom[0], data2.acom[0])

        self.assertApprox(self.data.com[0], data3.com[0])

        self.assertApprox(self.data.com[0], data4.com[0])
        self.assertApprox(self.data.vcom[0], data4.vcom[0])

    def test_com_default(self):
        v = rand(self.model.nv)
        a = rand(self.model.nv)
        pin.centerOfMass(self.model, self.data, self.q, v, a)

        data2 = self.model.createData()
        pin.forwardKinematics(self.model, data2, self.q, v, a)
        pin.centerOfMass(self.model, data2)

        for i in range(self.model.njoints):
            self.assertApprox(self.data.com[i], data2.com[i])
            self.assertApprox(self.data.vcom[i], data2.vcom[i])
            self.assertApprox(self.data.acom[i], data2.acom[i])

    def test_Jcom_update3(self):
        Jcom = pin.jacobianCenterOfMass(self.model, self.data, self.q)
        self.assertFalse(np.isnan(Jcom).any())

    def test_Jcom_update4(self):
        Jcom = pin.jacobianCenterOfMass(self.model, self.data, self.q, True)
        self.assertFalse(np.isnan(Jcom).any())
        self.assertFalse(np.isnan(self.data.com[1]).any())

    def test_Jcom_noupdate2(self):
        data_no = self.data
        data_up = self.model.createData()

        pin.forwardKinematics(self.model, data_no, self.q)
        Jcom_no = pin.jacobianCenterOfMass(self.model, data_no)

        Jcom_up = pin.jacobianCenterOfMass(self.model, data_up, self.q)

        self.assertTrue((Jcom_no == Jcom_up).all())

    def test_Jcom_noupdate3(self):
        data_no = self.data
        data_up = self.model.createData()

        pin.forwardKinematics(self.model, data_no, self.q)
        Jcom_no = pin.jacobianCenterOfMass(self.model, data_no, True)

        Jcom_up = pin.jacobianCenterOfMass(self.model, data_up, self.q, True)

        self.assertTrue((Jcom_no == Jcom_up).all())
        self.assertTrue((data_no.com[1] == data_up.com[1]).all())

    def test_subtree_jacobian(self):
        model = self.model
        data = self.data

        Jcom = pin.jacobianCenterOfMass(model, data, self.q)
        Jcom_subtree = pin.getJacobianSubtreeCenterOfMass(model, data, 0)
        self.assertApprox(Jcom, Jcom_subtree)

        data2 = model.createData()
        Jcom_subtree2 = pin.jacobianSubtreeCenterOfMass(model, data2, self.q, 0)
        self.assertApprox(Jcom_subtree, Jcom_subtree2)

        data3 = model.createData()
        Jcom_subtree3 = pin.jacobianSubtreeCenterOfMass(model, data3, self.q, 0)
        self.assertApprox(Jcom_subtree3, Jcom_subtree2)

        Jcom_subtree4 = pin.jacobianSubtreeCenterOfMass(model, data3, 0)
        self.assertApprox(Jcom_subtree3, Jcom_subtree4)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pinocchio as pin
from numpy.random import rand
from pinocchio import skew, skewSquare, unSkew

try:
    import casadi
    from pinocchio import casadi as cpin

    WITH_CASADI = True
except ImportError:
    WITH_CASADI = False

from test_case import PinocchioTestCase


class TestSpatial(PinocchioTestCase):
    def test_skew(self):
        v3 = rand(3)
        self.assertApprox(v3, unSkew(skew(v3)))
        self.assertLess(np.linalg.norm(skew(v3).dot(v3)), 1e-10)

        x, y, z = tuple(rand(3).tolist())
        M = np.array([[0.0, x, y], [-x, 0.0, z], [-y, -z, 0.0]])
        self.assertApprox(M, skew(unSkew(M)))

        rhs = rand(3)
        self.assertApprox(np.cross(v3, rhs, axis=0), skew(v3).dot(rhs))
        self.assertApprox(M.dot(rhs), np.cross(unSkew(M), rhs, axis=0))

        x, y, z = tuple(v3.tolist())
        self.assertApprox(skew(v3), np.array([[0, -z, y], [z, 0, -x], [-y, x, 0]]))

    def skewSquare(self):
        v3 = rand(3)

        Mss = skewSquare(v3)

        Ms = skew(v3)
        Mss_ref = Ms.dot(Ms)

        self.assertApprox(Mss, Mss_ref)

    def test_NaN_log3_casadi(self):
        if WITH_CASADI:
            Rtarget = pin.utils.rotate("x", 3.14 / 4)  # Target
            R0 = pin.Quaternion(0.707107, 0.707107, 0, 0).matrix()
            nu0 = pin.log3(R0)

            # Casadi symbolic variables and functions functions
            nu = casadi.SX.sym("v", 3, 1)
            D = cpin.log3(
                cpin.exp3(nu).T @ Rtarget
            )  # This seems to be the probelematic function
            dDi_dnu = [
                casadi.Function("gradient" + str(i), [nu], [casadi.gradient(D[i], nu)])
                for i in range(3)
            ]  # Compute the gradient of function D wrt nu

            d0 = dDi_dnu[0](nu0)  # Evaluate the gradient at a problematic point nu0
            self.assertFalse(
                np.any(np.isnan(d0)), "NaN detected in the log3 function derivative"
            )

    def test_Jlog6(self):
        for _ in range(10):
            M0: pin.SE3 = pin.SE3.Random()
            M1 = M0.copy()
            dM = M0.actInv(M1)
            J = pin.Jlog6(dM)
            R = dM.rotation
            logR = pin.log3(R)
            Jrot = pin.Jlog3(R)
            print(R, f":\nlog: {logR}")
            print(Jrot)

            self.assertApprox(dM, pin.SE3.Identity())
            self.assertApprox(Jrot, np.eye(3))
            self.assertApprox(J, np.eye(6))


if __name__ == "__main__":
    unittest.main()



import time
from pathlib import Path

try:
    import hppfcl
except ImportError:
    raise ImportError("hppfcl not found, but it is currently required by this viewer.")

import numpy as np

from .. import pinocchio_pywrap_default as pin
from . import BaseVisualizer

try:
    import trimesh  # Required by viser
    import viser
except ImportError:
    import_viser_succeed = False
else:
    import_viser_succeed = True


MESH_TYPES = (hppfcl.BVHModelBase, hppfcl.HeightFieldOBBRSS, hppfcl.HeightFieldAABB)


class ViserVisualizer(BaseVisualizer):
    """A Pinocchio visualizer using Viser."""

    def __init__(
        self,
        model=pin.Model(),
        collision_model=None,
        visual_model=None,
        copy_models=False,
        data=None,
        collision_data=None,
        visual_data=None,
    ):
        if not import_viser_succeed:
            msg = (
                "Error while importing the viewer client.\n"
                "Check whether viser is properly installed "
                "(pip install --user viser)."
            )
            raise ImportError(msg)

        super().__init__(
            model,
            collision_model,
            visual_model,
            copy_models,
            data,
            collision_data,
            visual_data,
        )
        self.static_objects = []

    def initViewer(
        self,
        viewer=None,
        open=False,
        loadModel=False,
        host="localhost",
        port="8000",
    ):
        """
        Start a new Viser server and client.
        Note: the server can also be passed in through the `viewer` argument.

        Parameters:
            viewer: An existing ViserServer instance, or None.
                If None, creates a new ViserServer in this visualizer.
            open: If True, automatically opens a browser tab to the visualizer.
            loadModel: If True, loads the Pinocchio models passed in.
            host: If `viewer` is None, this will be the host URL.
            port: If `viewer` is None, this will be the host port.
        """
        if (viewer is not None) and (not isinstance(viewer, viser.ViserServer)):
            raise RuntimeError(
                "'viewer' argument must be None or a valid ViserServer instance."
            )

        self.viewer = viewer or viser.ViserServer(host=host, server_port=port)
        self.frames = {}

        if open:
            import webbrowser

            webbrowser.open(f"http://{self.viewer.get_host()}:{self.viewer.get_port()}")

            # Wait until clients are reported.
            # Otherwise, capturing an image too soon after opening a browser window
            # may not register any clients.
            while len(self.viewer.get_clients()) == 0:
                time.sleep(0.1)

        if loadModel:
            self.loadViewerModel()

    def loadViewerModel(
        self,
        rootNodeName="pinocchio",
        collision_color=None,
        visual_color=None,
        frame_axis_length=0.2,
        frame_axis_radius=0.01,
    ):
        """Load the robot in a Viser viewer.

        Parameters:
            rootNodeName: name to give to the robot in the viewer
            collision_color: optional, color to give to the collision model of
                the robot. Format is a list of four RGBA floating-point numbers
                (between 0 and 1)
            visual_color: optional, color to give to the visual model of
                the robot. Format is a list of four RGBA floating-point numbers
                (between 0 and 1)
            frame_axis_length: optional, length of frame axes if displaying frames.
            frame_axis_radius: optional, radius of frame axes if displaying frames.
        """
        self.viewerRootNodeName = rootNodeName

        # Create root frames to help toggle visibility in the Viser UI.
        self.visualRootNodeName = rootNodeName + "/visual"
        self.visualRootFrame = self.viewer.scene.add_frame(
            self.visualRootNodeName, show_axes=False
        )
        self.collisionRootNodeName = rootNodeName + "/collision"
        self.collisionRootFrame = self.viewer.scene.add_frame(
            self.collisionRootNodeName, show_axes=False
        )
        self.framesRootNodeName = rootNodeName + "/frames"
        self.framesRootFrame = self.viewer.scene.add_frame(
            self.framesRootNodeName, show_axes=False
        )

        # Load visual model
        if (visual_color is not None) and (len(visual_color) != 4):
            raise RuntimeError("visual_color must have 4 elements for RGBA.")
        if self.visual_model is not None:
            for visual in self.visual_model.geometryObjects:
                self.loadViewerGeometryObject(
                    visual, self.visualRootNodeName, visual_color
                )
        self.displayVisuals(True)

        # Load collision model
        if (collision_color is not None) and (len(collision_color) != 4):
            raise RuntimeError("collision_color must have 4 elements for RGBA.")
        if self.collision_model is not None:
            for collision in self.collision_model.geometryObjects:
                self.loadViewerGeometryObject(
                    collision, self.collisionRootNodeName, collision_color
                )
        self.displayCollisions(False)

        # Load frames
        for frame in self.model.frames:
            frame_name = self.framesRootNodeName + "/" + frame.name
            self.frames[frame_name] = self.viewer.scene.add_frame(
                frame_name,
                show_axes=True,
                axes_length=frame_axis_length,
                axes_radius=frame_axis_radius,
            )
        self.displayFrames(False)

    def loadViewerGeometryObject(self, geometry_object, prefix="", color=None):
        """Loads a single geometry object."""
        name = geometry_object.name
        if prefix:
            name = prefix + "/" + name
        geom = geometry_object.geometry
        color_override = color or geometry_object.meshColor

        if isinstance(geom, hppfcl.Box):
            frame = self.viewer.scene.add_box(
                name,
                dimensions=geom.halfSide * 2.0,
                color=color_override[:3],
                opacity=color_override[3],
            )
        elif isinstance(geom, hppfcl.Sphere):
            frame = self.viewer.scene.add_icosphere(
                name,
                radius=geom.radius,
                color=color_override[:3],
                opacity=color_override[3],
            )
        elif isinstance(geom, hppfcl.Cylinder):
            mesh = trimesh.creation.cylinder(
                radius=geom.radius,
                height=geom.halfLength * 2.0,
            )
            frame = self.viewer.scene.add_mesh_simple(
                name,
                mesh.vertices,
                mesh.faces,
                color=color_override[:3],
                opacity=color_override[3],
            )
        elif isinstance(geom, MESH_TYPES):
            frame = self._add_mesh_from_path(
                name, geometry_object.meshPath, color_override
            )
        elif isinstance(geom, hppfcl.Convex):
            if len(geometry_object.meshPath) > 0:
                frame = self._add_mesh_from_path(
                    name, geometry_object.meshPath, color_override
                )
            else:
                frame = self._add_mesh_from_convex(name, geom, color_override)
        else:
            raise RuntimeError(f"Unsupported geometry type for {name}: {type(geom)}")

        self.frames[name] = frame

    def _add_mesh_from_path(self, name, mesh_path, color):
        """
        Load a mesh from a file.
        """
        extension = Path(mesh_path).suffix.lower()
        if extension == ".dae" or color is None:
            mesh = trimesh.load_scene(mesh_path)
            return self.viewer.scene.add_mesh_trimesh(name, mesh)
        else:
            mesh = trimesh.load_mesh(mesh_path)
            return self.viewer.scene.add_mesh_simple(
                name,
                mesh.vertices,
                mesh.faces,
                color=color[:3],
                opacity=color[3],
            )

    def _add_mesh_from_convex(self, name, geom, color):
        """
        Load a mesh from triangles stored inside a hppfcl.Convex.
        """
        num_tris = geom.num_polygons
        call_triangles = geom.polygons
        call_vertices = geom.points

        vertices = call_vertices()
        vertices = vertices.astype(np.float32)
        faces = np.empty((num_tris, 3), dtype=int)
        for k in range(num_tris):
            tri = call_triangles(k)
            faces[k] = [tri[i] for i in range(3)]

        return self.viewer.scene.add_mesh_simple(
            name,
            vertices,
            faces,
            color=color[:3],
            opacity=color[3],
        )

    def display(self, q=None):
        """
        Display the robot at configuration q in the viewer by placing all the bodies
        """
        if q is not None:
            pin.forwardKinematics(self.model, self.data, q)

        if self.collisionRootFrame.visible:
            self.updatePlacements(pin.GeometryType.COLLISION)

        if self.visualRootFrame.visible:
            self.updatePlacements(pin.GeometryType.VISUAL)

        if self.framesRootFrame.visible:
            self.updateFrames()

    def displayCollisions(self, visibility):
        self.collisionRootFrame.visible = visibility
        self.updatePlacements(pin.GeometryType.COLLISION)

    def displayVisuals(self, visibility):
        self.visualRootFrame.visible = visibility
        self.updatePlacements(pin.GeometryType.VISUAL)

    def displayFrames(self, visibility):
        self.framesRootFrame.visible = visibility
        self.updateFrames()

    def drawFrameVelocities(self, *args, **kwargs):
        raise NotImplementedError("drawFrameVelocities is not yet implemented.")

    def updatePlacements(self, geometry_type):
        if geometry_type == pin.GeometryType.VISUAL:
            geom_model = self.visual_model
            geom_data = self.visual_data
            prefix = self.viewerRootNodeName + "/visual"
        else:
            geom_model = self.collision_model
            geom_data = self.collision_data
            prefix = self.viewerRootNodeName + "/collision"

        pin.updateGeometryPlacements(self.model, self.data, geom_model, geom_data)
        for geom_id, geometry_object in enumerate(geom_model.geometryObjects):
            # Get mesh pose.
            M = geom_data.oMg[geom_id]

            # Update viewer configuration.
            frame_name = prefix + "/" + geometry_object.name
            frame = self.frames[frame_name]
            frame.position = M.translation * geometry_object.meshScale
            frame.wxyz = pin.Quaternion(M.rotation).coeffs()[
                [3, 0, 1, 2]
            ]  # Pinocchio uses xyzw

    def updateFrames(self):
        pin.updateFramePlacements(self.model, self.data)
        for frame_id, frame in enumerate(self.model.frames):
            # Get frame pose.
            M = self.data.oMf[frame_id]

            # Update viewer configuration.
            viser_frame_name = self.framesRootNodeName + "/" + frame.name
            viser_frame = self.frames[viser_frame_name]
            viser_frame.position = M.translation
            viser_frame.wxyz = pin.Quaternion(M.rotation).coeffs()[
                [3, 0, 1, 2]
            ]  # Pinocchio uses xyzw

    def captureImage(self, w=None, h=None, client_id=None, transport_format="jpeg"):
        """
        Capture an image from the Viser viewer and return an RGB array.

        Parameters:
            w: The width of the captured image. If None, uses the actual camera width.
            h: The height of the captured image. If None, uses the actual camera height.
            client_id: The ID of the Viser client handle.
                If None, uses the first available client.
            transport_format: The transport format to use for the captured image.
                Can be "jpeg" (default) or "png".
        """
        clients = self.viewer.get_clients()
        if len(clients) == 0:
            raise RuntimeError("Viser server has no attached clients!")

        if client_id is None:
            cli = next(iter(clients.values()))
        elif client_id not in clients:
            raise RuntimeError(
                f"Viser server does not have a client with ID '{client_id}'"
            )
        else:
            cli = clients[client_id]

        height = h or cli.camera.image_height
        width = w or cli.camera.image_width
        return cli.get_render(
            height=height, width=width, transport_format=transport_format
        )

    def setBackgroundColor(self, preset_name: str = "gray", col_top=None, col_bot=None):
        raise NotImplementedError("setBackgroundColor is not yet implemented.")

    def setCameraTarget(self, target: np.ndarray):
        raise NotImplementedError("setCameraTarget is not yet implemented.")

    def setCameraPosition(self, position: np.ndarray):
        raise NotImplementedError("setCameraPosition is not yet implemented.")

    def setCameraZoom(self, zoom: float):
        raise NotImplementedError("setCameraZoom is not yet implemented.")

    def setCameraPose(self, pose: np.ndarray = np.eye(4)):
        raise NotImplementedError("setcameraPose is not yet implemented.")

    def disableCameraControl(self):
        raise NotImplementedError("disableCameraControl is not yet implemented.")

    def enableCameraControl(self):
        raise NotImplementedError("enableCameraControl is not yet implemented.")



# This examples shows how to perform collision detection between the end-effector of a
# robot and a point cloud depicted as a Height Field
# Note: this feature requires Meshcat to be installed, this can be done using
# pip install --user meshcat

import os
import sys
import time
from pathlib import Path

import hppfcl as fcl
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

# Load the URDF model.

model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "panda.urdf"
urdf_model_path = model_path / "panda_description/urdf" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir
)

# Add point clouds
num_points = 5000
points = np.random.rand(3, num_points)
point_cloud_placement = (
    pin.SE3.Identity()
)  # Placement of the point cloud wrt the WORLD frame
point_cloud_placement.translation = np.array([0.2, 0.2, -0.5])

X = points[0, :]
Y = points[1, :]
Z = points[2, :]

nx = 20
x_grid = np.linspace(0.0, 1.0, nx)
x_half_pad = 0.5 * (x_grid[1] - x_grid[0])
x_bins = np.digitize(X, x_grid + x_half_pad)
x_dim = x_grid[-1] - x_grid[0]

ny = 20
y_grid = np.linspace(0.0, 1.0, ny)
y_half_pad = 0.5 * (y_grid[1] - y_grid[0])
y_bins = np.digitize(Y, y_grid + y_half_pad)
y_dim = y_grid[-1] - y_grid[0]

point_bins = y_bins * nx + x_bins
heights = np.zeros((ny, nx))
np.maximum.at(heights.ravel(), point_bins, Z)

point_cloud = fcl.BVHModelOBBRSS()
point_cloud.beginModel(0, num_points)
point_cloud.addVertices(points.T)

height_field = fcl.HeightFieldOBBRSS(x_dim, y_dim, heights, min(Z))
height_field_placement = point_cloud_placement * pin.SE3(
    np.eye(3), 0.5 * np.array([x_grid[0] + x_grid[-1], y_grid[0] + y_grid[-1], 0.0])
)

go_point_cloud = pin.GeometryObject(
    "point_cloud", 0, point_cloud_placement, point_cloud
)
go_point_cloud.meshColor = np.ones(4)
collision_model.addGeometryObject(go_point_cloud)
visual_model.addGeometryObject(go_point_cloud)

go_height_field = pin.GeometryObject(
    "height_field", 0, height_field_placement, height_field
)
go_height_field.meshColor = np.ones(4)
height_field_collision_id = collision_model.addGeometryObject(go_height_field)
visual_model.addGeometryObject(go_height_field)

# Add colllision pair between the height field and the panda_hand geometry
panda_hand_collision_id = collision_model.getGeometryId("panda_hand_0")
go_panda_hand = collision_model.geometryObjects[panda_hand_collision_id]
go_panda_hand.geometry.buildConvexRepresentation(False)
go_panda_hand.geometry = (
    go_panda_hand.geometry.convex
)  # We need to work with the convex hull of the real mesh

collision_pair = pin.CollisionPair(height_field_collision_id, panda_hand_collision_id)
collision_model.addCollisionPair(collision_pair)

# Start a new MeshCat server and client.
# Note: the server can also be started separately using the "meshcat-server" command in
# a terminal:
# this enables the server to remain active after the current script ends.
#
# Option open=True pens the visualizer.
# Note: the visualizer can also be opened seperately by visiting the provided URL.
try:
    viz = MeshcatVisualizer(model, collision_model, visual_model)
    viz.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

# Load the robot in the viewer.
viz.loadViewerModel()

# Display a robot configuration.
q0 = pin.neutral(model)
viz.display(q0)

is_collision = False
data = model.createData()
collision_data = collision_model.createData()
while not is_collision:
    q = pin.randomConfiguration(model)

    is_collision = pin.computeCollisions(
        model, data, collision_model, collision_data, q, True
    )

print("Found a configuration in collision:", q)
viz.display(q)
time.sleep(1.0)



import sys

import numpy as np
import pinocchio as pin

try:
    import hppfcl
except ImportError:
    print("This example requires hppfcl")
    sys.exit(0)
from pinocchio.visualize import MeshcatVisualizer

model = pin.Model()

geom_model = pin.GeometryModel()
geometries = [
    hppfcl.Capsule(0.1, 0.8),
    hppfcl.Sphere(0.5),
    hppfcl.Box(1, 1, 1),
    hppfcl.Cylinder(0.1, 1.0),
    hppfcl.Cone(0.5, 1.0),
]
for i, geom in enumerate(geometries):
    placement = pin.SE3(np.eye(3), np.array([i, 0, 0]))
    geom_obj = pin.GeometryObject(f"obj{i}", 0, 0, geom, placement)
    color = np.random.uniform(0, 1, 4)
    color[3] = 1
    geom_obj.meshColor = color
    geom_model.addGeometryObject(geom_obj)

viz = MeshcatVisualizer(model, geom_model, geom_model)

# Initialize the viewer.
try:
    viz.initViewer(open=True)
except ImportError as error:
    print(error)
    sys.exit(0)

try:
    viz.loadViewerModel("shapes")
except AttributeError as error:
    print(error)
    sys.exit(0)

viz.display(np.zeros(0))
input("press enter to continue")



import numpy as np
import pinocchio as pin

model = pin.buildSampleModelHumanoid()
model.lowerPositionLimit[:7] = -np.ones(7)
model.upperPositionLimit[:7] = +np.ones(7)

pool = pin.ModelPool(model)

num_threads = pin.omp_get_max_threads()
batch_size = 128
q = np.empty((model.nq, batch_size))
for k in range(batch_size):
    q[:, k] = pin.randomConfiguration(model)

v = np.zeros((model.nv, batch_size))
a = np.zeros((model.nv, batch_size))
tau = np.zeros((model.nv, batch_size))

print(f"num_threads: {num_threads}")
print(f"batch_size: {batch_size}")

# Call RNEA
res_rnea = np.empty((model.nv, batch_size))
pin.rneaInParallel(num_threads, pool, q, v, a, res_rnea)  # Without allocation
res_rnea2 = pin.rneaInParallel(num_threads, pool, q, v, a)  # With allocation

# Call ABA
res_aba = np.empty((model.nv, batch_size))
pin.abaInParallel(num_threads, pool, q, v, tau, res_aba)  # Without allocation
res_aba2 = pin.abaInParallel(num_threads, pool, q, v, tau)  # With allocation



#
# Copyright (c) 2016 CNRS
#
# ruff: noqa: E731

import numpy as np
from pinocchio import Force, Motion, skew
from pinocchio.utils import zero


def jFromIdx(idxv, robot):
    """Return the joint index from the velocity index"""
    for j in range(1, robot.model.njoint):
        if idxv in range(
            robot.model.joints[j].idx_v,
            robot.model.joints[j].idx_v + robot.model.joints[j].nv,
        ):
            return j


parent = lambda i, robot: robot.model.parents[i]
iv = lambda i, robot: list(
    range(
        robot.model.joints[i].idx_v,
        robot.model.joints[i].idx_v + robot.model.joints[i].nv,
    )
)
ancestors = lambda j, robot, res=[]: (
    res
    if j == 0
    else ancestors(
        robot.model.parents[j],
        robot,
        [j, *res],
    )
)


class ancestorOf:
    def __init__(self, i, robot):
        self.dec = i
        self.robot = robot

    def __contains__(self, anc):
        dec = self.dec
        while dec > 0:
            if anc == dec:
                return True
            else:
                dec = self.robot.model.parents[dec]


# descendants = lambda root,robot: filter(
# lambda i: root in ancestorOf(i,robot), range(root,robot.model.njoints) )
descendants = lambda root, robot: robot.model.subtrees[root]


def setRobotArgs(robot):
    ancestors.__defaults__ = (robot, *ancestors.__defaults__)
    descendants.__defaults__ = (robot,)
    # ancestorsOf.__init__.__defaults__ = (robot,)
    iv.__defaults__ = (robot,)
    parent.__defaults__ = (robot,)
    jFromIdx.__defaults__ = (robot,)


# --- SE3 operators
Mcross = lambda x, y: Motion(x).cross(Motion(y)).vector
Mcross.__doc__ = "Motion cross product"

Fcross = lambda x, y: Motion(x).cross(Force(y)).vector
Fcross.__doc__ = "Force cross product"

MCross = lambda V, v: np.bmat([Mcross(V[:, i], v) for i in range(V.shape[1])])
FCross = lambda V, f: np.bmat([Fcross(V[:, i], f) for i in range(V.shape[1])])


adj = lambda nu: np.bmat([[skew(nu[3:]), skew(nu[:3])], [zero([3, 3]), skew(nu[3:])]])
adj.__doc__ = "Motion pre-cross product (ie adjoint, lie bracket operator)"

adjdual = lambda nu: np.bmat(
    [[skew(nu[3:]), zero([3, 3])], [skew(nu[:3]), skew(nu[3:])]]
)
adjdual.__doc__ = "Force pre-cross product adjdual(a) = -adj(a)' "

td = np.tensordot
quad = lambda H, v: np.array(td(td(H, v, [2, 0]), v, [1, 0]))
quad.__doc__ = """Tensor product v'*H*v, with H.shape = [ nop, nv, nv ]"""


def np_prettyprint(sarg="{: 0.5f}", eps=5e-7):
    mformat = lambda x, sarg=sarg, eps=eps: (
        sarg.format(x) if abs(x) > eps else " 0.     "
    )
    np.set_printoptions(formatter={"float": mformat})



import contextlib
import os
from pathlib import Path


def get_dll_paths():
    pinocchio_paths = os.getenv("PINOCCHIO_WINDOWS_DLL_PATH")
    if pinocchio_paths is None:
        # Standard site-packages to bin path
        RELATIVE_DLL_PATH = "..\\..\\..\\bin"
        return [Path(__file__).parent / RELATIVE_DLL_PATH]
    else:
        return pinocchio_paths.split(os.pathsep)


class PathManager(contextlib.AbstractContextManager):
    """Restore PATH state after importing Python module"""

    def add_dll_directory(self, dll_dir: str):
        os.environ["PATH"] += os.pathsep + dll_dir

    def __enter__(self):
        self.old_path = os.environ["PATH"]
        return self

    def __exit__(self, *exc_details):
        os.environ["PATH"] = self.old_path


class DllDirectoryManager(contextlib.AbstractContextManager):
    """Restore DllDirectory state after importing Python module"""

    def add_dll_directory(self, dll_dir: str):
        # add_dll_directory can fail on relative path and non
        # existing path.
        # Since we don't know all the fail criterion we just ignore
        # thrown exception
        try:
            self.dll_dirs.append(os.add_dll_directory(dll_dir))
        except OSError:
            pass

    def __enter__(self):
        self.dll_dirs = []
        return self

    def __exit__(self, *exc_details):
        for d in self.dll_dirs:
            d.close()


def build_directory_manager():
    return DllDirectoryManager()



import os
from pathlib import Path

import pinocchio as pin

model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "romeo_small.urdf"
urdf_model_path = model_path / "romeo_description/urdf" / urdf_filename

# Load model
model = pin.buildModelFromUrdf(urdf_model_path, pin.JointModelFreeFlyer())

# Load collision geometries
geom_model = pin.buildGeomFromUrdf(
    model, urdf_model_path, pin.GeometryType.COLLISION, mesh_dir
)

# Add collisition pairs
geom_model.addAllCollisionPairs()
print("num collision pairs - initial:", len(geom_model.collisionPairs))

# Remove collision pairs listed in the SRDF file
srdf_filename = "romeo.srdf"
srdf_model_path = model_path / "romeo_description/srdf" / srdf_filename

pin.removeCollisionPairs(model, geom_model, srdf_model_path)
print(
    "num collision pairs - after removing useless collision pairs:",
    len(geom_model.collisionPairs),
)

# Load reference configuration
pin.loadReferenceConfigurations(model, srdf_model_path)

# Retrieve the half sitting position from the SRDF file
q = model.referenceConfigurations["half_sitting"]

# Create data structures
data = model.createData()
geom_data = pin.GeometryData(geom_model)


# Compute all the collisions
pin.computeCollisions(model, data, geom_model, geom_data, q, False)

# Print the status of collision for all collision pairs
for k in range(len(geom_model.collisionPairs)):
    cr = geom_data.collisionResults[k]
    cp = geom_model.collisionPairs[k]
    print(
        "collision pair:",
        cp.first,
        ",",
        cp.second,
        "- collision:",
        "Yes" if cr.isCollision() else "No",
    )

# Compute for a single pair of collision
pin.updateGeometryPlacements(model, data, geom_model, geom_data, q)
pin.computeCollision(geom_model, geom_data, 0)



"""
Deep actor-critic network,
From "Continuous control with deep reinforcement learning",
by Lillicrap et al, arXiv:1509.02971
"""

import random
import signal
import time
from collections import deque

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tflearn
from pendulum import Pendulum

# --- Random seed
RANDOM_SEED = int((time.time() % 10) * 1000)
print(f"Seed = {RANDOM_SEED}")
np.random.seed(RANDOM_SEED)
tf.set_random_seed(RANDOM_SEED)
random.seed(RANDOM_SEED)
n_init = tflearn.initializations.truncated_normal(seed=RANDOM_SEED)
u_init = tflearn.initializations.uniform(minval=-0.003, maxval=0.003, seed=RANDOM_SEED)

# --- Hyper paramaters
NEPISODES = 100  # Max training steps
NSTEPS = 100  # Max episode length
QVALUE_LEARNING_RATE = 0.001  # Base learning rate for the Q-value Network
POLICY_LEARNING_RATE = 0.0001  # Base learning rate for the policy network
DECAY_RATE = 0.99  # Discount factor
UPDATE_RATE = 0.01  # Homotopy rate to update the networks
REPLAY_SIZE = 10000  # Size of replay buffer
BATCH_SIZE = 64  # Number of points to be fed in stochastic gradient
NH1 = NH2 = 250  # Hidden layer size

# --- Environment
env = Pendulum(1)  # Continuous pendulum
env.withSinCos = True  # State is dim-3: (cosq,sinq,qdot) ...
NX = env.nobs  # ... training converges with q,qdot with 2x more neurones.
NU = env.nu  # Control is dim-1: joint torque

# --- Q-value and policy networks


class QValueNetwork:
    def __init__(self):
        nvars = len(tf.trainable_variables())

        x = tflearn.input_data(shape=[None, NX])
        u = tflearn.input_data(shape=[None, NU])

        netx1 = tflearn.fully_connected(x, NH1, weights_init=n_init, activation="relu")
        netx2 = tflearn.fully_connected(netx1, NH2, weights_init=n_init)
        netu1 = tflearn.fully_connected(
            u, NH1, weights_init=n_init, activation="linear"
        )
        netu2 = tflearn.fully_connected(netu1, NH2, weights_init=n_init)
        net = tflearn.activation(netx2 + netu2, activation="relu")
        qvalue = tflearn.fully_connected(net, 1, weights_init=u_init)

        self.x = x  # Network state   <x> input in Q(x,u)
        self.u = u  # Network control <u> input in Q(x,u)
        self.qvalue = qvalue  # Network output  <Q>
        # Variables to be trained
        self.variables = tf.trainable_variables()[nvars:]
        self.hidens = [netx1, netx2, netu1, netu2]  # Hidden layers for debug

    def setupOptim(self):
        qref = tf.placeholder(tf.float32, [None, 1])
        loss = tflearn.mean_square(qref, self.qvalue)
        optim = tf.train.AdamOptimizer(QVALUE_LEARNING_RATE).minimize(loss)
        gradient = tf.gradients(self.qvalue, self.u)[0] / float(BATCH_SIZE)

        self.qref = qref  # Reference Q-values
        self.optim = optim  # Optimizer
        self.gradient = (
            # Gradient of Q wrt the control  dQ/du (for policy training)
            gradient
        )
        return self

    def setupTargetAssign(self, nominalNet, tau=UPDATE_RATE):
        self.update_variables = [
            target.assign(tau * ref + (1 - tau) * target)
            for target, ref in zip(self.variables, nominalNet.variables)
        ]
        return self


class PolicyNetwork:
    def __init__(self):
        nvars = len(tf.trainable_variables())

        x = tflearn.input_data(shape=[None, NX])
        net = tflearn.fully_connected(x, NH1, activation="relu", weights_init=n_init)
        net = tflearn.fully_connected(net, NH2, activation="relu", weights_init=n_init)
        policy = (
            tflearn.fully_connected(net, NU, activation="tanh", weights_init=u_init)
            * env.umax
        )

        self.x = x  # Network input <x> in Pi(x)
        self.policy = policy  # Network output <Pi>
        # Variables to be trained
        self.variables = tf.trainable_variables()[nvars:]

    def setupOptim(self):
        qgradient = tf.placeholder(tf.float32, [None, NU])
        grad = tf.gradients(self.policy, self.variables, -qgradient)
        optim = tf.train.AdamOptimizer(POLICY_LEARNING_RATE).apply_gradients(
            zip(grad, self.variables)
        )

        # Q-value gradient wrt control (input value)
        self.qgradient = qgradient
        self.optim = optim  # Optimizer
        return self

    def setupTargetAssign(self, nominalNet, tau=UPDATE_RATE):
        self.update_variables = [
            target.assign(tau * ref + (1 - tau) * target)
            for target, ref in zip(self.variables, nominalNet.variables)
        ]
        return self


# --- Replay memory
class ReplayItem:
    def __init__(self, x, u, r, d, x2):
        self.x = x
        self.u = u
        self.reward = r
        self.done = d
        self.x2 = x2


replayDeque = deque()

# --- Tensor flow initialization

policy = PolicyNetwork().setupOptim()
policyTarget = PolicyNetwork().setupTargetAssign(policy)

qvalue = QValueNetwork().setupOptim()
qvalueTarget = QValueNetwork().setupTargetAssign(qvalue)

sess = tf.InteractiveSession()
tf.global_variables_initializer().run()

# Uncomment to save or restore networks
# tf.train.Saver().restore(sess, "netvalues/actorcritic.pre.ckpt")
# tf.train.Saver().save   (sess, "netvalues/actorcritic.full.ckpt")


def rendertrial(maxiter=NSTEPS, verbose=True):
    x = env.reset()
    rsum = 0.0
    for i in range(maxiter):
        u = sess.run(policy.policy, feed_dict={policy.x: x.T})
        x, reward = env.step(u)
        env.render()
        time.sleep(1e-2)
        rsum += reward
    if verbose:
        print("Lasted ", i, " timestep -- total reward:", rsum)


signal.signal(
    signal.SIGTSTP, lambda x, y: rendertrial()
)  # Roll-out when CTRL-Z is pressed

# History of search
h_rwd = []
h_qva = []
h_ste = []

# --- Training
for episode in range(1, NEPISODES):
    x = env.reset().T
    rsum = 0.0

    for step in range(NSTEPS):
        # Greedy policy ...
        u = sess.run(policy.policy, feed_dict={policy.x: x})
        u += 1.0 / (1.0 + episode + step)  # ... with noise
        x2, r = env.step(u)
        x2 = x2.T
        done = False  # pendulum scenario is endless.

        # Feed replay memory ...
        replayDeque.append(ReplayItem(x, u, r, done, x2))
        if len(replayDeque) > REPLAY_SIZE:
            replayDeque.popleft()  # ... with FIFO forgetting.

        rsum += r
        if done or np.linalg.norm(x - x2) < 1e-3:
            break  # Break when pendulum is still.
        x = x2

        # Start optimizing networks when memory size > batch size.
        if len(replayDeque) > BATCH_SIZE:
            batch = random.sample(
                replayDeque, BATCH_SIZE
            )  # Random batch from replay memory.
            x_batch = np.vstack([b.x for b in batch])
            u_batch = np.vstack([b.u for b in batch])
            r_batch = np.vstack([b.reward for b in batch])
            d_batch = np.vstack([b.done for b in batch])
            x2_batch = np.vstack([b.x2 for b in batch])

            # Compute Q(x,u) from target network
            u2_batch = sess.run(
                policyTarget.policy, feed_dict={policyTarget.x: x2_batch}
            )
            q2_batch = sess.run(
                qvalueTarget.qvalue,
                feed_dict={qvalueTarget.x: x2_batch, qvalueTarget.u: u2_batch},
            )
            qref_batch = r_batch + (not d_batch) * (DECAY_RATE * q2_batch)

            # Update qvalue to solve HJB constraint: q = r + q'
            sess.run(
                qvalue.optim,
                feed_dict={
                    qvalue.x: x_batch,
                    qvalue.u: u_batch,
                    qvalue.qref: qref_batch,
                },
            )

            # Compute approximate policy gradient ...
            u_targ = sess.run(policy.policy, feed_dict={policy.x: x_batch})
            qgrad = sess.run(
                qvalue.gradient, feed_dict={qvalue.x: x_batch, qvalue.u: u_targ}
            )
            # ... and take an optimization step along this gradient.
            sess.run(
                policy.optim, feed_dict={policy.x: x_batch, policy.qgradient: qgrad}
            )

            # Update target networks by homotopy.
            sess.run(policyTarget.update_variables)
            sess.run(qvalueTarget.update_variables)

    # \\\END_FOR step in range(NSTEPS)

    # Display and logging (not mandatory).
    maxq = (
        np.max(
            sess.run(qvalue.qvalue, feed_dict={qvalue.x: x_batch, qvalue.u: u_batch})
        )
        if "x_batch" in locals()
        else 0
    )
    print(
        f"Ep#{episode:3d}: lasted {step:d} steps, "
        f"reward={rsum:3.0f}, max qvalue={maxq:2.3f}"
    )
    h_rwd.append(rsum)
    h_qva.append(maxq)
    h_ste.append(step)
    if not (episode + 1) % 20:
        rendertrial(100)

# \\\END_FOR episode in range(NEPISODES)

print(f"Average reward during trials: {sum(h_rwd) / NEPISODES:.3f}")
rendertrial()
plt.plot(np.cumsum(h_rwd) / range(1, NEPISODES))
plt.show()



import unittest

import pinocchio as pin


class TestJointCompositeBindings(unittest.TestCase):
    def test_basic(self):
        jc = pin.JointModelComposite()
        self.assertTrue(hasattr(jc, "joints"))
        self.assertTrue(hasattr(jc, "njoints"))
        self.assertTrue(hasattr(jc, "jointPlacements"))

    def test_empty_constructor(self):
        jc = pin.JointModelComposite()
        self.assertTrue(jc.nq == 0)
        self.assertTrue(len(jc.joints) == 0)
        self.assertTrue(jc.njoints == len(jc.joints))

    def test_reserve_constructor(self):
        jc = pin.JointModelComposite(2)
        self.assertTrue(jc.nq == 0)
        self.assertTrue(len(jc.joints) == 0)
        self.assertTrue(jc.njoints == len(jc.joints))

    def test_joint_constructor(self):
        j1 = pin.JointModelRX()
        self.assertTrue(j1.nq == 1)

        jc1 = pin.JointModelComposite(j1)
        self.assertTrue(jc1.nq == 1)
        self.assertTrue(len(jc1.joints) == 1)
        self.assertTrue(jc1.njoints == len(jc1.joints))

        j2 = pin.JointModelRX()
        self.assertTrue(j2.nq == 1)

        jc2 = pin.JointModelComposite(j1, pin.SE3.Random())
        self.assertTrue(jc2.nq == 1)
        self.assertTrue(len(jc2.joints) == 1)
        self.assertTrue(jc2.njoints == len(jc2.joints))

    def test_add_joint(self):
        j1 = pin.JointModelRX()
        self.assertTrue(j1.nq == 1)
        j2 = pin.JointModelRY()
        self.assertTrue(j2.nq == 1)
        j3 = pin.JointModelRZ()
        self.assertTrue(j3.nq == 1)

        jc = pin.JointModelComposite(2)
        self.assertTrue(jc.nq == 0)
        self.assertTrue(len(jc.joints) == 0)
        self.assertTrue(jc.njoints == len(jc.joints))

        jc.addJoint(j1)
        self.assertTrue(jc.nq == 1)
        self.assertTrue(len(jc.joints) == 1)
        self.assertTrue(jc.njoints == len(jc.joints))

        jc.addJoint(j2)
        self.assertTrue(jc.nq == 2)
        self.assertTrue(len(jc.joints) == 2)
        self.assertTrue(jc.njoints == len(jc.joints))

        jc.addJoint(j3, pin.SE3.Random())
        self.assertTrue(jc.nq == 3)
        self.assertTrue(jc.njoints == len(jc.joints))

    def test_add_joint_return(self):
        jc1 = pin.JointModelComposite()
        jc2 = jc1.addJoint(pin.JointModelRX())
        jc3 = jc2.addJoint(pin.JointModelRY())
        jc4 = jc1.addJoint(pin.JointModelRZ())
        self.assertTrue(jc1.njoints == 3)
        self.assertTrue(jc2.njoints == 3)
        self.assertTrue(jc3.njoints == 3)
        self.assertTrue(jc4.njoints == 3)

        del jc1
        del jc3
        del jc4
        self.assertTrue(jc2.njoints == 3)

    def test_add_joint_concat(self):
        j1 = pin.JointModelRX()
        self.assertTrue(j1.nq == 1)
        j2 = pin.JointModelRY()
        self.assertTrue(j2.nq == 1)
        j3 = pin.JointModelRZ()
        self.assertTrue(j3.nq == 1)

        jc = pin.JointModelComposite(j1).addJoint(j2, pin.SE3.Random()).addJoint(j3)

        self.assertTrue(jc.nq == 3)
        self.assertTrue(len(jc.joints) == 3)
        self.assertTrue(jc.njoints == len(jc.joints))


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Inria
# SPDX-License-Identifier: BSD-2-Clause


"""
In this short script, we show how to compute inverse dynamics (RNEA), i.e. the
vector of joint torques corresponding to a given motion.
"""

from pathlib import Path

import numpy as np
import pinocchio as pin

# Load the model from a URDF file
# Change to your own URDF file here, or give a path as command-line argument
pinocchio_model_dir = Path(__file__).parent.parent / "models/"
model_path = pinocchio_model_dir / "example-robot-data/robots"
mesh_dir = pinocchio_model_dir
urdf_filename = "ur5_robot.urdf"
urdf_model_path = model_path / "ur_description/urdf/" / urdf_filename
model, _, _ = pin.buildModelsFromUrdf(urdf_model_path, package_dirs=mesh_dir)

# Build a data frame associated with the model
data = model.createData()

# Sample a random joint configuration, joint velocities and accelerations
q = pin.randomConfiguration(model)  # in rad for the UR5
v = np.random.rand(model.nv, 1)  # in rad/s for the UR5
a = np.random.rand(model.nv, 1)  # in rad/s² for the UR5

# Computes the inverse dynamics (RNEA) for all the joints of the robot
tau = pin.rnea(model, data, q, v, a)

# Print out to the vector of joint torques (in N.m)
print("Joint torques: " + str(tau))



import sys

import casadi
import hppfcl as fcl
import numpy as np
import pinocchio as pin
import pinocchio.casadi as cpin
from pinocchio.visualize import MeshcatVisualizer


def make_cartpole(ub=True):
    model = pin.Model()

    m1 = 1.0
    m2 = 0.1
    length = 0.5
    base_sizes = (0.4, 0.2, 0.05)

    base = pin.JointModelPX()
    base_id = model.addJoint(0, base, pin.SE3.Identity(), "base")

    if ub:
        pole = pin.JointModelRUBY()
    else:
        pole = pin.JointModelRY()
    pole_id = model.addJoint(1, pole, pin.SE3.Identity(), "pole")

    base_inertia = pin.Inertia.FromBox(m1, *base_sizes)
    pole_inertia = pin.Inertia(
        m2,
        np.array([0.0, 0.0, length / 2]),
        m2 / 5 * np.diagflat([1e-2, length**2, 1e-2]),
    )

    base_body_pl = pin.SE3.Identity()
    pole_body_pl = pin.SE3.Identity()
    pole_body_pl.translation = np.array([0.0, 0.0, length / 2])

    model.appendBodyToJoint(base_id, base_inertia, base_body_pl)
    model.appendBodyToJoint(pole_id, pole_inertia, pole_body_pl)

    # make visual/collision models
    collision_model = pin.GeometryModel()
    shape_base = fcl.Box(*base_sizes)
    radius = 0.01
    shape_pole = fcl.Capsule(radius, length)
    RED_COLOR = np.array([1, 0.0, 0.0, 1.0])
    WHITE_COLOR = np.array([1, 1.0, 1.0, 1.0])
    geom_base = pin.GeometryObject("link_base", base_id, shape_base, base_body_pl)
    geom_base.meshColor = WHITE_COLOR
    geom_pole = pin.GeometryObject("link_pole", pole_id, shape_pole, pole_body_pl)
    geom_pole.meshColor = RED_COLOR

    collision_model.addGeometryObject(geom_base)
    collision_model.addGeometryObject(geom_pole)
    visual_model = collision_model
    return model, collision_model, visual_model


class PinocchioCasadi:
    """Take a Pinocchio model, turn it into a Casadi model
    and define the appropriate graphs.
    """

    def __init__(self, model: pin.Model, timestep=0.05):
        self.model = model
        self.cmodel = cpin.Model(model)  # cast to CasADi model
        self.cdata = self.cmodel.createData()
        self.timestep = timestep
        self.create_dynamics()
        self.create_discrete_dynamics()

    def create_dynamics(self):
        """Create the acceleration expression and acceleration function."""
        nq = self.model.nq
        nu = 1
        nv = self.model.nv
        q = casadi.SX.sym("q", nq)
        v = casadi.SX.sym("v", nv)
        u = casadi.SX.sym("u", nu)
        dq_ = casadi.SX.sym("dq_", nv)
        self.u_node = u
        self.q_node = q
        self.v_node = v
        self.dq_ = dq_

        B = np.array([1, 0])
        tau = B @ u
        a = cpin.aba(self.cmodel, self.cdata, q, v, tau)
        self.acc = a
        self.acc_func = casadi.Function("acc", [q, v, u], [a], ["q", "v", "u"], ["a"])

    def create_discrete_dynamics(self):
        """
        Create the map `(q,v) -> (qnext, vnext)` using semi-implicit Euler integration.
        """
        q = self.q_node
        v = self.v_node
        u = self.u_node
        dq_ = self.dq_
        # q' = q + dq
        q_dq = cpin.integrate(self.cmodel, q, dq_)
        self.q_dq = q_dq
        # express acceleration using q' = q + dq
        a = self.acc_func(q_dq, v, u)

        dt = self.timestep
        vnext = v + a * dt
        qnext = cpin.integrate(self.cmodel, self.q_dq, dt * vnext)

        self.dyn_qv_fn_ = casadi.Function(
            "discrete_dyn",
            [q, dq_, v, u],
            [qnext, vnext],
            ["q", "dq_", "v", "u"],
            ["qnext", "vnext"],
        )

    def forward(self, x, u):
        nq = self.model.nq
        nv = self.model.nv
        q = x[:nq]
        v = x[nq:]
        dq_ = np.zeros(nv)
        qnext, vnext = self.dyn_qv_fn_(q, dq_, v, u)
        xnext = np.concatenate((qnext, vnext))
        return xnext

    def residual_fwd(self, x, u, xnext):
        nv = self.model.nv
        dq = np.zeros(nv)
        dqn = dq
        res = self.dyn_residual(x, u, xnext, dq, dqn)
        return res


class CartpoleDynamics(PinocchioCasadi):
    def __init__(self, timestep=0.05):
        model, collision_model, visual_model = make_cartpole()
        self.collision_model = collision_model
        self.visual_model = visual_model
        super().__init__(model=model, timestep=timestep)


dt = 0.02
cartpole = CartpoleDynamics(timestep=dt)
model = cartpole.model

print(model)

q0 = np.array([0.0, 0.95, 0.01])
q0 = pin.normalize(model, q0)
v = np.zeros(model.nv)
u = np.zeros(1)
a0 = cartpole.acc_func(q0, v, u)

print("a0:", a0)

x0 = np.append(q0, v)
xnext = cartpole.forward(x0, u)


def integrate_no_control(x0, nsteps):
    states_ = [x0.copy()]
    for t in range(nsteps):
        u = np.zeros(1)
        xnext = cartpole.forward(states_[t], u).ravel()
        states_.append(xnext)
    return states_


states_ = integrate_no_control(x0, nsteps=400)
states_ = np.stack(states_).T

try:
    viz = MeshcatVisualizer(
        model=model,
        collision_model=cartpole.collision_model,
        visual_model=cartpole.visual_model,
    )

    viz.initViewer()
    viz.loadViewerModel("pinocchio")

    qs_ = states_[: model.nq, :].T
    viz.play(q_trajectory=qs_, dt=dt)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)



#
# Copyright (c) 2015-2021 CNRS INRIA
#
# ruff: noqa: E402, F401, F403, F405

import numpy

# On Windows, if pinocchio.dll is not in the same directory than
# the .pyd, it will not be loaded.
# We first try to load pinocchio, then, if it fail and we are on Windows:
#  1. We add all paths inside PINOCCHIO_WINDOWS_DLL_PATH to DllDirectory
#  2. If PINOCCHIO_WINDOWS_DLL_PATH we add the relative path from the
#     package directory to the bin directory to DllDirectory
# This solution is inspired from:
#  - https://github.com/PixarAnimationStudios/OpenUSD/pull/1511/files
#  - https://stackoverflow.com/questions/65334494/python-c-extension-packaging-dll-along-with-pyd
# More resources on https://github.com/diffpy/pyobjcryst/issues/33
try:
    from .pinocchio_pywrap_default import *
    from .pinocchio_pywrap_default import __raw_version__, __version__
except ImportError:
    import platform

    if platform.system() == "Windows":
        from .windows_dll_manager import build_directory_manager, get_dll_paths

        with build_directory_manager() as dll_dir_manager:
            for p in get_dll_paths():
                dll_dir_manager.add_dll_directory(p)
            from .pinocchio_pywrap_default import *
            from .pinocchio_pywrap_default import __raw_version__, __version__
    else:
        raise

import inspect

# Manually register submodules
import sys

from . import utils
from .explog import exp, log

submodules = inspect.getmembers(pinocchio_pywrap_default, inspect.ismodule)
for module_info in submodules:
    sys.modules["pinocchio." + module_info[0]] = module_info[1]

sys.modules["pinocchio.rpy"] = rpy
sys.modules["pinocchio.cholesky"] = cholesky

if WITH_HPP_FCL:
    try:
        import hppfcl
        from hppfcl import (
            CachedMeshLoader,
            CollisionGeometry,
            CollisionResult,
            Contact,
            DistanceResult,
            MeshLoader,
            StdVec_CollisionResult,
            StdVec_Contact,
            StdVec_DistanceResult,
        )

        WITH_HPP_FCL_BINDINGS = True
    except ImportError:
        WITH_HPP_FCL_BINDINGS = False
else:
    WITH_HPP_FCL_BINDINGS = False

from .deprecated import *
from .robot_wrapper import RobotWrapper
from .shortcuts import *
